# flake8: noqa

# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

# Import relevant classes for easy access
from speakeasy.windows.common import PeFile
from speakeasy.windows.win32 import Win32Emulator
from speakeasy.windows.kernel import WinKernelEmulator
from speakeasy.speakeasy import Speakeasy

import speakeasy.winenv.defs as defs
import speakeasy.winenv.arch as arch


# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import unittest

import util


class TestWdmEmulation(unittest.TestCase):

    def setUp(self):
        self.config = util.get_config()
        self.dev_name = '\\Device\\wdm_test'
        self.sym_link = '\\DosDevices\\wdm_test'
        self.report = None

    def _get_api_calls(self, ep, api_name):
        return [api for api in ep['apis'] if api['api_name'] == api_name]

    def _test_driver_load_unload(self):
        eps = self.report['entry_points']

        driver_entry = eps[0]

        create_dev = self._get_api_calls(driver_entry, 'ntoskrnl.IoCreateDeviceSecure')
        self.assertEqual(1, len(create_dev))
        create_dev = create_dev[0]
        self.assertEqual(create_dev['args'][2], self.dev_name)

        create_sym = self._get_api_calls(driver_entry, 'ntoskrnl.IoCreateSymbolicLink')
        self.assertEqual(1, len(create_sym))
        create_sym = create_sym[0]
        self.assertEqual(create_sym['args'][0], self.sym_link)
        self.assertEqual(create_sym['args'][1], self.dev_name)

        self.assertEqual(driver_entry['ret_val'], '0x0')

        driver_unload = eps[-1]
        delete_sym = self._get_api_calls(driver_unload, 'ntoskrnl.IoDeleteSymbolicLink')
        self.assertEqual(1, len(delete_sym))
        delete_sym = delete_sym[0]
        self.assertEqual(delete_sym['args'][0], self.sym_link)

        delete_dev = self._get_api_calls(driver_unload, 'ntoskrnl.IoDeleteDevice')
        self.assertEqual(1, len(delete_dev))
        delete_dev = delete_dev[0]
        self.assertNotEqual(delete_dev['args'][0], "0x0")

    def _test_irp_handlers(self):
        eps = self.report['entry_points']
        irp_handlers = [ep for ep in eps if ep['ep_type'].startswith('irp_')]
        self.assertEqual(len(irp_handlers), 6)
        for ih in irp_handlers:
            if ih['ep_type'] == 'irp_mj_create':
                dprint = self._get_api_calls(ih, 'ntoskrnl.DbgPrint')
                self.assertEqual(1, len(dprint))
                dprint = dprint[0]
                self.assertEqual(dprint['args'][0], 'Inside IRP_MJ_CREATE handler')
                self.assertEqual(ih['ret_val'], '0x0')
            elif ih['ep_type'] == 'irp_mj_device_control':
                dprint = self._get_api_calls(ih, 'ntoskrnl.DbgPrint')
                self.assertEqual(1, len(dprint))
                dprint = dprint[0]
                self.assertEqual(dprint['args'][0], 'Inside IRP_MJ_DEVICE_CONTROL handler')
                self.assertEqual(ih['ret_val'], '0x0')
            elif ih['ep_type'] == 'irp_mj_close':
                dprint = self._get_api_calls(ih, 'ntoskrnl.DbgPrint')
                self.assertEqual(1, len(dprint))
                dprint = dprint[0]
                self.assertEqual(dprint['args'][0], 'Inside IRP_MJ_CLOSE handler')
                self.assertEqual(ih['ret_val'], '0x0')
            else:
                dprint = self._get_api_calls(ih, 'ntoskrnl.DbgPrint')
                self.assertEqual(1, len(dprint))
                dprint = dprint[0]
                self.assertEqual(dprint['args'][0], 'Inside default handler')
                self.assertEqual(ih['ret_val'], '0xc00000bb')

    def _test_emu_wdm_driver(self, fpath):
        data = util.get_test_bin_data(fpath)
        self.report = util.run_test(self.config, data)

    def test_x86_wdm_driver_entry(self):
        self._test_emu_wdm_driver('wdm_test_x86.sys.xz')
        self._test_driver_load_unload()

    def test_x64_wdm_driver_entry(self):
        self._test_emu_wdm_driver('wdm_test_x64.sys.xz')
        self._test_driver_load_unload()

    def test_x86_wdm_irp_handlers(self):
        self._test_emu_wdm_driver('wdm_test_x86.sys.xz')
        self._test_irp_handlers()

    def test_x64_wdm_irp_handlers(self):
        self._test_emu_wdm_driver('wdm_test_x64.sys.xz')
        self._test_irp_handlers()


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import os
import ntpath

import speakeasy.winenv.defs.nt.ntoskrnl as ntoskrnl
import speakeasy.winenv.arch as _arch
import speakeasy.winenv.defs.nt.ddk as ddk
import speakeasy.winenv.defs.windows.windows as windef


class Console(object):
    """
    Represents a console window object
    """
    curr_handle = 0x340

    def __init__(self):
        self.handle = self.get_handle()
        self.window = 0

    def get_handle(self):
        tmp = Console.curr_handle
        Console.curr_handle += 4
        return tmp

    def set_window(self, window):
        self.window = window

    def get_window(self):
        return self.window


class SEH(object):
    """
    Implements the structures needed to support SEH handling during emulation
    """
    class ScopeRecord(object):
        def __init__(self, record):
            self.record = record
            self.filter_called = False
            self.handler_called = False

    class Frame(object):

        def __init__(self, entry, scope_table, scope_records):
            self.entry = entry
            self.scope_table = scope_table
            self.scope_records = []
            for rec in scope_records:
                SEH.ScopeRecord(rec)
                self.scope_records.append(SEH.ScopeRecord(rec))
            self.searched = False

    def __init__(self):
        self.context = None
        self.context_address = 0
        self.record = None
        self.frames = []
        self.last_func = 0
        self.last_exception_code = 0
        self.exception_ptrs = 0
        self.handler_ret_val = None

    def set_context(self, context, address=0):
        self.context = context
        self.context_address = address

    def get_context(self):
        return self.context

    def set_last_func(self, func):
        self.last_func = func

    def set_record(self, record, address=0):
        self.record = record

    def set_current_frame(self, frame):
        self.frame = frame

    def get_frames(self):
        return self.frames

    def clear_frames(self):
        self.frames = []

    def add_frame(self, entry, scope_table, records):
        frame = SEH.Frame(entry, scope_table, records)
        self.frames.append(frame)


class KernelObject(object):
    """
    Base class for Kernel objects managed by the object manager
    """
    curr_handle = 0x220
    curr_id = 0x400

    def __init__(self, emu):
        self.emu = emu
        self.address = None
        self.name = ''
        self.object = 0
        self.ref_cnt = 0
        self.handles = []
        self.arch = emu.get_arch()
        self.id = KernelObject.curr_id
        KernelObject.curr_id += 4

        self.nt_types = ntoskrnl
        self.win_types = windef

    def sizeof(self, obj=None):
        if obj:
            return obj.sizeof()
        return self.object.sizeof()

    def get_bytes(self, obj=None):
        if obj:
            return obj.get_bytes()
        return self.object.get_bytes()

    def read_back(self):
        data = self.emu.mem_read(self.address, self.sizeof())
        self.object.cast(data)
        return self

    def write_back(self):
        data = self.get_bytes()
        if data and self.address:
            self.emu.mem_write(self.address, data)

    def get_id(self):
        return self.id

    def set_id(self, oid):
        self.id = oid

    def get_class_name(self):
        if self.object:
            return self.object.__class__.__name__

    def get_mem_tag(self):
        return 'emu.struct.%s' % (self.get_class_name())

    def get_handle(self):
        tmp = KernelObject.curr_handle
        KernelObject.curr_handle += 4
        self.handles.append(tmp)
        return tmp


class Driver(KernelObject):

    """
    Class that represents DRIVER_OBJECTs created by the Windows kernel
    """

    ldr_entries = []

    def __init__(self, emu):
        super(Driver, self).__init__(emu=emu)
        self.pe = None
        self.devices = []
        self.mj_funcs = [None] * (ddk.IRP_MJ_MAXIMUM_FUNCTION + 1)
        self.on_unload = None
        self.unload_called = False
        self.reg_path_ptr = 0
        self.reg_path = ''
        self.name = ''
        self.basename = ''
        self.object = self.nt_types.DRIVER_OBJECT(emu.get_ptr_size())

    def create_reg_path(self, name):
        """
        Create the service path in the registry for the created driver
        """
        self.reg_path = '\\Registry\\Machine\\System\\CurrentControlSet\\Services\\%s\x00' % (name) # noqa
        buf = self.reg_path.encode('utf-16le')
        self.reg_path = self.reg_path.strip('\x00')

        us = self.nt_types.UNICODE_STRING(self.emu.get_ptr_size())
        size = self.sizeof(us) + len(buf)
        addr = self.emu.mem_map(size, tag='emu.object.%s.reg_path' % self.name)
        us.Length = len(buf) - 2
        us.MaximumLength = len(buf)
        us.Buffer = addr + self.sizeof(us)

        self.emu.mem_write(addr, self.get_bytes(us))
        self.emu.mem_write(us.Buffer, buf)

        self.reg_path_ptr = addr

    def get_basename(self):
        return self.basename.lower()

    def get_reg_path(self):
        return self.reg_path

    def init_driver_section(self):
        """
        Create the driver section for the driver. This is a linked list that
        links together driver objects
        """
        tag = 'emu.object.%s.DriverSection' % self.name
        mod = self.pe

        if mod and mod.get_emu_path():
            mod_name = mod.get_emu_path()
        else:
            mod_name = 'None'

        ldte = LdrDataTableEntry(self.emu, mod_name, tag=tag)
        first = None
        last = None
        if len(self.ldr_entries):
            first = self.ldr_entries[0]
            last = self.ldr_entries[-1]
        self.ldr_entries.append(ldte)

        ldte.object.DllBase = mod.get_base()

        dllname = mod.get_emu_path().encode('utf-16le')
        name_addr = ldte.address + ldte.sizeof()
        self.emu.mem_write(name_addr, dllname)

        # Set the dll full name
        ldte.object.FullDllName.Length = len(dllname)
        ldte.object.FullDllName.MaximumLength = len(dllname)
        ldte.object.FullDllName.Buffer = name_addr

        # Set the dll base name
        dllname = ntpath.basename(mod.get_emu_path()).encode('utf-16le')
        ldte.object.BaseDllName.Length = len(dllname)
        ldte.object.BaseDllName.MaximumLength = len(dllname)
        ldte.object.BaseDllName.Buffer = name_addr + \
            (ldte.object.FullDllName.Length - len(dllname))

        if not last:
            ldte.object.InLoadOrderLinks.Flink = ldte.address
            ldte.object.InLoadOrderLinks.Flink = ldte.address
            ldte.object.InLoadOrderLinks.Blink = ldte.address
            ldte.object.InMemoryOrderLinks.Blink = ldte.address
        else:
            ldte.object.InLoadOrderLinks.Flink = \
                last.object.InLoadOrderLinks.Flink
            ldte.object.InMemoryOrderLinks.Flink = \
                last.object.InMemoryOrderLinks.Flink
            ldte.object.InLoadOrderLinks.Blink = last.address
            ldte.object.InMemoryOrderLinks.Blink = last.address

            last.object.InLoadOrderLinks.Flink = ldte.address
            last.object.InMemoryOrderLinks.Flink = ldte.address
            first.object.InLoadOrderLinks.Blink = ldte.address
            first.object.InMemoryOrderLinks.Blink = ldte.address

            last.write_back()
        ldte.write_back()
        return ldte

    def init_driver_object(self, name=None, pe=None, is_decoy=True):
        """
        Initialize the DRIVER_OBJECT
        """
        self.pe = pe
        drvobj = self.object

        drvobj.Type = 4
        drvobj.Size = self.sizeof()
        drvobj.DeviceObject = 0
        drvobj.Flags = 2
        us = ''
        if pe:
            drvobj.DriverStart = pe.get_base()
            drvobj.DriverSize = pe.image_size
            drvobj.DriverInit = pe.get_base() + pe.ep

            if is_decoy:
                ep = pe.get_base() + pe.ep
                drvobj.MajorFunction[ddk.IRP_MJ_CREATE] = ep + 1
                drvobj.MajorFunction[ddk.IRP_MJ_READ] = ep + 2
                drvobj.MajorFunction[ddk.IRP_MJ_WRITE] = ep + 3
                drvobj.MajorFunction[ddk.IRP_MJ_DEVICE_CONTROL] = ep + 4
                drvobj.MajorFunction[ddk.IRP_MJ_PNP] = ep + 5
                drvobj.MajorFunction[ddk.IRP_MJ_INTERNAL_DEVICE_CONTROL] = ep+6

            if not name:
                drvname = os.path.splitext(pe.path)[0]
                drvname = os.path.basename(drvname)
                self.name = r'\Driver\%s' % (drvname)
                us = ('\\Driver\\%s\x00' % (drvname)).encode('utf-16le')
                name = self.name

        if name:
            us = name.encode('utf-16le')
            self.name = name

        if not pe and not self.name:
            name = 'none'

        # Allocate the driver object
        addr = self.emu.mem_map(drvobj.Size + len(us),
                                tag='emu.object.%s' % name)

        drvobj.DriverName.Length = len(us)
        drvobj.DriverName.MaximumLength = len(us)
        drvobj.DriverName.Buffer = addr + drvobj.Size

        name = self.name
        idx = name.rfind('\\')
        if idx >= 0 and idx != len(name) - 1:
            name = self.name[idx+1:]
        else:
            name = 'None'

        self.basename = name
        self.create_reg_path(name)

        drv_sect = self.init_driver_section()
        drvobj.DriverSection = drv_sect.address

        self.address = addr
        self.emu.mem_write(addr, self.get_bytes() + us)

    def read_back(self):
        super(Driver, self).read_back()

        for i, func in enumerate(self.mj_funcs):
            self.mj_funcs[i] = self.object.MajorFunction[i]

        self.on_unload = self.object.DriverUnload


class Device(KernelObject):
    """
    Represents a DEVICE_OBJECT created by the windows kernel
    """

    def __init__(self, emu):
        super(Device, self).__init__(emu=emu)
        devobj = self.nt_types.DEVICE_OBJECT(emu.get_ptr_size())

        devobj.Type = 0x3
        devobj.Size = self.sizeof(devobj)
        devobj.ReferenceCount = 1

        self.file_object = None
        self.object = devobj
        self.driver = None

    def get_parent_driver(self):
        return self.driver


class FileObject(KernelObject):
    """
    Represents a FILE_OBJECT created by the windows kernel
    """
    def __init__(self, emu):
        super(FileObject, self).__init__(emu=emu)
        fileobj = self.nt_types.FILE_OBJECT(emu.get_ptr_size())

        fileobj.Type = 0x5
        fileobj.Size = self.sizeof(fileobj)
        self.object = fileobj
        self.address = emu.mem_map(self.sizeof(), tag=self.get_mem_tag())


class IoStackLocation(KernelObject):
    """
    Represents a IO_STACK_LOCATION struct that is part of
    an IRP.
    """
    def __init__(self, emu):
        super(IoStackLocation, self).__init__(emu=emu)

        self.object = self.nt_types.IO_STACK_LOCATION(emu.get_ptr_size())
        # Allocate two stack locations for now to handle IoGetNextIrpStackLocation calls
        self.address = emu.mem_map(self.sizeof() * 2, tag=self.get_mem_tag())


class Irp(KernelObject):
    """
    I/O request packet used when performing device input/output
    """
    def __init__(self, emu):
        super(Irp, self).__init__(emu=emu)
        self.object = self.nt_types.IRP(emu.get_ptr_size())
        self.address = emu.mem_map(self.sizeof(), tag=self.get_mem_tag())
        self.stack_locations = []

        ios = IoStackLocation(emu=emu)
        ios.write_back()

        self.object.Tail.Overlay.CurrentStackLocation = ios.address + ios.sizeof()

        self.stack_locations.append(ios)

        self.object.Type = 0x6
        self.object.Size = self.sizeof()
        self.write_back()

    def get_curr_stack_loc(self):
        return self.stack_locations[0]


class Thread(KernelObject):
    """
    Represents a Windows ETHREAD object that describes a
    an OS level thread
    """
    def __init__(self, emu, stack_base=0, stack_commit=0):
        super(Thread, self).__init__(emu=emu)
        self.emu = emu
        self.object = self.nt_types.ETHREAD(emu.get_ptr_size())
        self.address = emu.mem_map(self.sizeof(), tag=self.get_mem_tag())
        self.object.Data = b'\xFF' * self.sizeof()
        self.tid = self.get_id()
        self.modified_pc = False
        self.teb = None
        self.seh = SEH()
        self.tls = []
        self.message_queue = []
        self.ctx = self.emu.get_thread_context()
        self.fls = []
        self.suspend_count = 0
        self.token = Token(self.emu)
        self.last_error = 0
        self.stack_base = stack_base
        self.stack_commit = stack_commit

        self.write_back()
        self.process = None

    def queue_message(self, msg):
        """
        Add a GUI message to the thread's message queue
        """
        self.message_queue.append(msg)

    def get_seh(self):
        return self.seh

    def get_context(self):
        if self.ctx:
            return self.ctx
        return self.emu.get_thread_context()

    def set_context(self, ctx):
        if self.ctx:
            if self.emu.get_arch() == _arch.ARCH_X86:
                if ctx.Eip != self.ctx.Eip:
                    self.modified_pc = True
        self.ctx = ctx

    def init_teb(self, teb_addr, peb_addr):
        if not self.teb:
            self.teb = TEB(emu=self.emu, address=teb_addr)

        self.teb.object.NtTib.StackBase = self.stack_base
        self.teb.object.NtTib.Self = teb_addr
        self.teb.object.NtTib.StackLimit = self.stack_commit
        self.teb.object.ProcessEnvironmentBlock = peb_addr
        self.teb.write_back()

    def get_teb(self):
        return self.teb.read_back()

    def set_last_error(self, code):
        self.last_error = code

    def get_last_error(self):
        return self.last_error

    def get_tls(self):
        return self.tls

    def set_tls(self, tls):
        self.tls = tls

    def get_fls(self):
        return self.fls

    def set_fls(self, fls):
        self.fls = fls

    def get_token(self):
        return self.token


class Token(KernelObject):
    """
    Represents a TOKEN object
    """
    def __init__(self, emu):
        super(Token, self).__init__(emu=emu)


class Process(KernelObject):

    """
    An EPROCESS object used by the Windows kernel to represent a process
    """
    ldr_entries = []

    def __init__(self, emu, pe=None, user_modules=[],
                 name='', path='', cmdline='', base=0, session=0):
        super(Process, self).__init__(emu=emu)
        # TODO: For now just allocate a blank opaque struct for an EPROCESS
        self.object = self.nt_types.EPROCESS(emu.get_ptr_size())
        self.address = emu.mem_map(self.sizeof(), tag=self.get_mem_tag(), perms=1, base=0xe0000000)
        self.name = name
        self.base = base
        self.pid = self.get_id()
        self.modules = user_modules
        self.threads = []
        self.console = None
        self.curr_thread = None
        self.cmdline = cmdline
        self.session = session
        self.token = Token(self.emu)
        emu.add_object(self.token)
        self.pe = pe

        self.stdin = (0xf000) + 1
        self.stdout = (0xf000) + 2
        self.stderr = (0xf000) + 3

        # Initialize the process PEB
        self.peb = PEB(emu=emu)
        self.peb_ldr_data = PebLdrData(self.emu)
        self.is_peb_active = False
        self.path = path
        self.set_process_parameters(emu)
        self.image = ''
        self.title = ''

        if pe and pe.OPTIONAL_HEADER.Subsystem & ddk.WINDOWS_CONSOLE:
            self.alloc_console()

    def get_peb(self):
        return self.peb

    def set_peb_ldr_address(self, addr):
        self.peb.object.Ldr = addr
        self.peb.write_back()
        self.peb_ldr_data.address = addr

    def set_process_parameters(self, emu):
        process_parameters = RTL_USER_PROCESS_PARAMETERS(emu=emu, proc=self)
        self.peb.object.ProcessParameters = process_parameters.address
        self.peb.write_back()

    def get_peb_ldr(self):
        return self.peb_ldr_data

    def alloc_console(self):

        if not self.console:
            self.console = Console()
        sm = self.emu.get_session_manager()
        desk = sm.get_current_desktop()
        self.console.set_window(desk.new_window())

    def get_desktop_name(self):
        sm = self.emu.get_session_manager()
        stat = sm.get_current_station()
        stat_name = stat.get_name()

        desk = sm.get_current_desktop()
        desk_name = desk.get_name()

        name = '%s\\%s' % (stat_name, desk_name)

        return name

    def get_token(self):
        """
        Get the token associated with the process
        """
        return self.token

    def get_std_handle(self, dev):
        STD_INPUT_HANDLE = 0xfffffff6
        STD_OUTPUT_HANDLE = 0xfffffff5
        STD_ERROR_HANDLE = 0xfffffff4

        for k, v in ((STD_INPUT_HANDLE, self.stdin),
                     (STD_OUTPUT_HANDLE, self.stdout),
                     (STD_ERROR_HANDLE, self.stderr),
                     ):

            if k == dev:
                return v
        return 0

    def get_title_name(self):
        return self.title

    def get_module(self):
        return self.pe

    def get_ep(self):
        if self.pe:
            return self.pe.get_ep()

    def get_console(self):
        return self.console

    def get_session_id(self):
        return self.session

    def get_pid(self):
        return self.pid

    def get_process_path(self):
        return self.path

    def get_command_line(self):
        return self.cmdline

    def set_user_modules(self, mods):
        self.modules = mods

    def new_thread(self):
        thr = Thread(self.emu)
        self.threads.append(thr)

    def add_module_to_peb(self, module):
        pld = self.peb_ldr_data
        list_type = self.nt_types.LIST_ENTRY(self.emu.get_ptr_size())

        # Initialize the LDTE
        ldte = LdrDataTableEntry(self.emu, module.get_emu_path())
        if not self.ldr_entries:
            prev = ldte
        else:
            prev = self.ldr_entries[-1]

        self.ldr_entries.append(ldte)
        first = self.ldr_entries[0]

        ldte.object.InLoadOrderLinks.Flink = first.address
        ldte.object.InMemoryOrderLinks.Flink = first.address + self.sizeof(list_type)
        ldte.object.InInitializationOrderLinks.Flink = first.address + self.sizeof(list_type) * 2

        ldte.object.DllBase = module.get_base()
        dllname = (module.get_emu_path() + '\x00').encode('utf-16le')
        name_addr = ldte.address + ldte.sizeof()
        self.emu.mem_write(name_addr, dllname)

        # Set the dll full name
        ldte.object.FullDllName.Length = len(dllname) - 2
        ldte.object.FullDllName.MaximumLength = len(dllname)
        ldte.object.FullDllName.Buffer = name_addr

        # Set the dll base name
        dllname = (ntpath.basename(module.get_emu_path()) +
                   '\x00').encode('utf-16le')

        ldte.object.BaseDllName.Length = len(dllname) - 2
        ldte.object.BaseDllName.MaximumLength = len(dllname)
        ldte.object.BaseDllName.Buffer = name_addr + \
            (ldte.object.FullDllName.MaximumLength - len(dllname))
        ldte.write_back()

        prev.object.InLoadOrderLinks.Flink = ldte.address
        prev.object.InMemoryOrderLinks.Flink = ldte.address + \
            self.sizeof(list_type)

        if first is ldte:
            prev.object.InInitializationOrderLinks.Flink = 0
        else:
            imol = prev.object.InMemoryOrderLinks.Flink
            prev.object.InInitializationOrderLinks.Flink = imol + \
                self.sizeof(list_type)

        ldte.object.InLoadOrderLinks.Blink = prev.address
        ldte.object.InMemoryOrderLinks.Blink = prev.address + \
            self.sizeof(list_type)

        if first is ldte:
            ldte.object.InInitializationOrderLinks.Blink = 0
        else:
            imol = ldte.object.InMemoryOrderLinks.Blink
            ldte.object.InInitializationOrderLinks.Blink = imol + \
                self.sizeof(list_type)

        prev.write_back()
        ldte.write_back()

        first.object.InLoadOrderLinks.Blink = prev.address
        first.object.InMemoryOrderLinks.Blink = prev.address
        if first is not ldte:
            first.object.InInitializationOrderLinks.Blink = prev.address

        first.write_back()

        pld.object.InLoadOrderModuleList.Flink = first.address
        pld.object.InMemoryOrderModuleList.Flink = \
            pld.object.InLoadOrderModuleList.Flink + \
            self.sizeof(list_type)

        # Lets just copy InMemoryOrderModuleList but skip the main EXE module
        head = pld.object.InMemoryOrderModuleList.Flink
        le = self.emu.mem_cast(ntoskrnl.LIST_ENTRY(self.emu.get_ptr_size()),
                               head)

        pld.object.InInitializationOrderModuleList.Flink = le.Flink + self.sizeof(list_type)

        pld.object.InLoadOrderModuleList.Blink = prev.address
        pld.object.InMemoryOrderModuleList.Blink = \
            pld.object.InLoadOrderModuleList.Blink + \
            self.sizeof(list_type)

        pld.object.InInitializationOrderModuleList.Blink = \
            pld.object.InMemoryOrderModuleList.Blink + \
            self.sizeof(list_type)

        pld.write_back()

        self.peb.object.Ldr = pld.address
        self.peb.write_back()

    def init_peb(self, modules):
        # Add an entry for each module in the module list
        for mod in modules:
            self.add_module_to_peb(mod)


class RTL_USER_PROCESS_PARAMETERS(KernelObject):
    def __init__(self, emu, proc):
        super(RTL_USER_PROCESS_PARAMETERS, self).__init__(emu=emu)

        self.object = self.nt_types.RTL_USER_PROCESS_PARAMETERS(emu.get_ptr_size())
        proc_path = (proc.path + '\x00').encode('utf-16le')
        proc_cmdline = (proc.cmdline + '\x00').encode('utf-16le')
        size = self.sizeof()
        size += len(proc_path)
        size += len(proc_cmdline)
        self.address = emu.mem_map(size,
                                   tag=proc.get_mem_tag() + '.ProcessParameters')
        emu.mem_write(self.address + self.sizeof(), proc_path)
        emu.mem_write(self.address + self.sizeof() + len(proc_path), proc_cmdline)

        self.object.ImagePathName.Length = len(proc_path) - 2
        self.object.ImagePathName.MaxLength = len(proc_path)
        self.object.ImagePathName.Buffer = self.address + self.sizeof()

        self.object.CommandLine.Length = len(proc_cmdline) - 2
        self.object.CommandLine.MaxLength = len(proc_cmdline)
        self.object.CommandLine.Buffer = self.address + self.sizeof() + len(proc_path)
        self.write_back()


class PEB(KernelObject):
    """
    Represents the process environment block. This structure contains a large amount of
    fields that are used internally by Windows. Shellcode may parse this structure in
    order to resolve exported functions.
    """
    def __init__(self, emu, address=None):
        super(PEB, self).__init__(emu=emu)

        self.object = self.nt_types.PEB(emu.get_ptr_size())
        if not address:
            self.address = emu.mem_map(self.sizeof(), tag=self.get_mem_tag())
        else:
            self.address = address


class TEB(KernelObject):
    """
    Represents the thread environment block. This structure contains a large amount of
    fields that are used internally by Windows.
    """
    def __init__(self, emu, address=0):
        super(TEB, self).__init__(emu=emu)

        self.object = self.nt_types.TEB(emu.get_ptr_size())
        if address:
            self.address = address
        else:
            self.address = emu.mem_map(self.sizeof(), tag=self.get_mem_tag())


class PebLdrData(KernelObject):

    def __init__(self, emu):
        super(PebLdrData, self).__init__(emu=emu)
        self.object = self.nt_types.PEB_LDR_DATA(emu.get_ptr_size())
        self.address = 0


class LdrDataTableEntry(KernelObject):
    def __init__(self, emu, dllname, tag=''):
        super(LdrDataTableEntry, self).__init__(emu=emu)
        self.object = self.nt_types.LDR_DATA_TABLE_ENTRY(emu.get_ptr_size())

        size = self.sizeof()
        size += len((dllname + '\x00').encode('utf-16le'))

        if not tag:
            tag = self.get_mem_tag()

        self.address = emu.mem_map(size, tag=tag)


class IDT(KernelObject):
    """
    Represents the Interrupt descriptor table. This is currently a dummy structure and
    only exists to detect if samples read or write to it.
    """
    def __init__(self, emu):
        super(IDT, self).__init__(emu=emu)

        self.object = self.nt_types.IDT(emu.get_ptr_size())
        self.address = emu.mem_map(self.sizeof(), tag=self.get_mem_tag())

    def init_descriptors(self):
        tbl = self.nt_types.DESCRIPTOR_TABLE(self.emu.get_ptr_size())

        kbase = self.emu.get_kernel_base()

        descs = self.emu.mem_map(self.sizeof(tbl),
                                 tag=self.get_mem_tag() + '.idt_entries')
        self.object.Limit = 0xFFF
        self.object.Descriptors = descs

        if self.emu.get_arch() == _arch.ARCH_X86:
            for i, entry in enumerate(tbl.Table):
                entry.OffsetLow = 0 + (4 * i)
                entry.Base = kbase
        elif self.emu.get_arch() == _arch.ARCH_AMD64:
            for i, entry in enumerate(tbl.Table):
                entry.OffsetLow = 0xFFFF & kbase
                entry.OffsetMiddle = (0xFFFF0000 & kbase) >> 16
                entry.OffsetHigh = (0xFFFFFFFF00000000 & kbase) >> 32

        self.emu.mem_write(descs, self.get_bytes(tbl))

        self.write_back()


class Event(KernelObject):
    """
    Describes event objects used by Windows for synchronization
    """

    def __init__(self, emu):
        super(Event, self).__init__(emu=emu)
        self.object = self.nt_types.KEVENT(emu.get_ptr_size())
        self.address = emu.mem_map(self.sizeof(), tag=self.get_mem_tag())


class Mutant(KernelObject):
    """
    Describes mutant objects used by Windows for synchronization
    """

    def __init__(self, emu):
        super(Mutant, self).__init__(emu=emu)
        self.object = self.nt_types.MUTANT(emu.get_ptr_size())
        self.address = emu.mem_map(self.sizeof(), tag=self.get_mem_tag())


class ObjectManager(object):
    """
    Class that manages kernel objects during emulation
    """
    def __init__(self, emu):
        super(ObjectManager, self).__init__()
        self.emu = emu
        self.objects = {}
        self.symlinks = []

    def add_symlink(self, link, dev):
        self.symlinks.append((link, dev))

    def new_object(self, obj_type):

        obj = obj_type(emu=self.emu)
        obj.set_id(self.new_id())
        return self.add_object(obj)

    def add_object(self, obj):
        if not self.get_object_from_addr(obj.address):
            self.objects.update({obj.address: obj})
        if not obj.id:
            obj.id = self.new_id()
        obj.ref_cnt += 1
        return obj

    def remove_object(self, obj):
        """
        Remove an object from the object manager
        """
        addr = None
        for a, o in self.objects.items():
            if o == obj:
                addr = a
                break
        if addr:
            self.objects.pop(addr)

    def dec_ref(self, obj):
        """
        Dereferece an object and remove it from the object manager when its reference count is 0
        """
        if hasattr(obj, 'ref_cnt'):
            obj.ref_cnt -= 1
            if obj.ref_cnt <= 0:
                self.remove_object(obj)
            return obj.ref_cnt

    def get_handle(self, obj):
        tmp = KernelObject.curr_handle
        KernelObject.curr_handle += 4
        obj.handles.append(tmp)
        return tmp

    def new_id(self):
        _id = KernelObject.curr_id
        KernelObject.curr_id += 4
        return _id

    def get_object_from_addr(self, addr):
        return self.objects.get(addr)

    def get_object_from_id(self, id):
        for a, o in self.objects.items():
            if o.get_id() == id:
                return o

    def get_object_from_name(self, name, check_symlinks=True):

        if not name:
            return None
        name = name.rstrip('\\')
        for a, o in self.objects.items():
            if not o.name:
                continue
            if o.name.lower() == name.lower():
                return o
        if check_symlinks:
            m = [sl[1] for sl in self.symlinks
                 if name.lower() == sl[0].lower()]
            if m:
                name = m[0]
            return self.get_object_from_name(name, False)

    def get_object_from_handle(self, handle):
        for a, o in self.objects.items():
            if handle in o.handles:
                return o



from .. import api

import time

class Winmm(api.ApiHandler):
    """
    Emulates functions from winmm.dll
    """

    name = 'winmm'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):
        
        super(Winmm, self).__init__(emu)
        super(Winmm, self).__get_hook_attrs__(self)

    @apihook('timeGetTime', argc=0)
    def timeGetTime(self, emu, argv, ctx={}):        
        '''
        DWORD timeGetTime(); // return the system time, in milliseconds
        '''
        return int(time.monotonic() * 1000) & 0xffffffff
        

        



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

RESOURCE_CONNECTED = 1
RESOURCE_GLOBALNET = 2
RESOURCE_REMEMBERED = 3
RESOURCE_CONTEXT = 5

RESOURCETYPE_ANY = 0
RESOURCETYPE_DISK = 1
RESOURCETYPE_PRINT = 2

RESOURCEUSAGE_CONNECTABLE = 1
RESOURCEUSAGE_CONTAINER = 2
RESOURCEUSAGE_ATTACHED = 0x10
RESOURCEUSAGE_ALL = 0x13

ERROR_NO_NETWORK = 0x4C6


def get_define_int(define, prefix=''):
    for k, v in globals().items():
        if not isinstance(v, int) or v != define:
            continue
        if prefix:
            if k.startswith(prefix):
                return k
        else:
            return k



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

class GuiObject(object):
    """
    Base class for all GUI objects
    """
    curr_handle = 0x120

    def __init__(self):
        self.handle = self.get_handle()

    def get_handle(self):
        tmp = GuiObject.curr_handle
        GuiObject.curr_handle += 4
        return tmp


class Session(GuiObject):
    """
    Represents a windows Session
    """
    def __init__(self, sess_id):
        super(Session, self).__init__()
        self.id = sess_id
        self.stations = {}

    def new_station(self, name='WinSta0'):
        stat = Station(name=name)
        self.stations.update({stat.get_handle(): stat})
        return stat


class Station(GuiObject):
    """
    Represents a window station
    """
    def __init__(self, name=''):
        super(Station, self).__init__()
        self.name = name
        self.desktops = {}

    def new_desktop(self, name=''):
        desk = Desktop(name=name)
        self.desktops.update({desk.get_handle(): desk})
        return desk

    def get_name(self):
        return self.name


class Desktop(GuiObject):
    """
    Represents a Desktop object
    """
    def __init__(self, name=''):
        super(Desktop, self).__init__()
        self.windows = {}
        self.desktop_window = self.new_window()
        self.name = name

    def new_window(self):
        # create the desktop window
        window = Window()
        self.windows.update({window.get_handle(): window})
        return window

    def get_desktop_window(self):
        return self.desktop_window

    def get_name(self):
        return self.name


class Window(GuiObject):
    """
    Represents a GUI window
    """
    def __init__(self, name=None, class_name=None):
        super(Window, self).__init__()
        self.name = name
        self.class_name = class_name


class WindowClass(GuiObject):
    """
    Represents a GUI window class
    """
    def __init__(self, wclass, name):
        super(WindowClass, self).__init__()
        self.wclass = wclass
        self.name = name


class SessionManager(object):
    """
    The session manager for the emulator. This will manage things like desktops,
    windows, and session isolation
    """
    def __init__(self, config):
        super(SessionManager, self).__init__()
        self.sessions = {}
        self.window_classes = {}
        self.windows = {}
        self.curr_session = None
        self.curr_station = None
        self.curr_desktop = None
        self.config = config
        self.dev_ctx = GuiObject.curr_handle

        # create a session 0
        self.curr_session = Session(sess_id=0)

        # create WinSta0
        self.curr_station = self.curr_session.new_station(name='WinSta0')

        # Create a desktop
        self.curr_station.new_desktop('Winlogon')
        default = self.curr_station.new_desktop('Default')
        self.curr_station.new_desktop('Disconnect')

        # For now lets default to the Default desktop
        self.curr_desktop = default

    def create_window_class(self, class_obj, class_name=None):
        wc = WindowClass(class_obj, class_name)
        atom = wc.get_handle()
        self.window_classes.update({atom: wc})
        if class_name:
            self.window_classes.update({class_name: wc})
        return atom

    def create_window(self, window_name=None, class_name=None):
        wc = Window(window_name, class_name)
        hnd = wc.get_handle()
        self.windows.update({hnd: wc})
        if window_name:
            self.windows.update({window_name: wc})
        return hnd

    def get_window_class(self, atom):
        return self.window_classes.get(atom)

    def get_window(self, handle):
        return self.windows.get(handle)

    def get_device_context(self):
        return self.dev_ctx

    def get_current_desktop(self):
        return self.curr_desktop

    def get_current_station(self):
        return self.curr_station

    def get_gui_object(self, handle):
        for hsess, sess in self.sessions.items():
            if hsess == handle:
                return sess
            for hstat, stat in sess.stations.items():
                if hstat == handle:
                    return stat
                for hdesk, desk in stat.desktops.items():
                    if hdesk == handle:
                        return desk



# Copyright (C) 2021 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.defs.windows.netapi32 as netapi32defs
import speakeasy.winenv.arch as _arch

from .. import api


class Wkscli(api.ApiHandler):
    """
    Implements exported functions from wkscli.dll
    """
    name = 'wkscli'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):
        super(Wkscli, self).__init__(emu)
        super(Wkscli, self).__get_hook_attrs__(self)

    @apihook('NetGetJoinInformation', argc=3)
    def NetGetJoinInformation(self, emu, argv, ctx={}):
        """
        NET_API_STATUS NET_API_FUNCTION NetGetJoinInformation(
          LPCWSTR lpServer,
          LPWSTR *lpNameBuffer,
          PNETSETUP_JOIN_STATUS BufferType
        );
        """
        lpServer, lpNameBuffer, BufferType = argv

        if lpServer:
            server = self.read_wide_string(lpServer)
            argv[0] = server

        # Assumes the server being queried is the local computer
        domain = emu.get_domain()
        argv[1] = domain
        namebuf = self.mem_alloc(emu.get_ptr_size())
        self.write_wide_string(domain, namebuf)
        self.mem_write(lpNameBuffer, namebuf.to_bytes(emu.get_ptr_size(), 'little'))

        argv[2] = netapi32defs.NetSetupDomainName
        self.mem_write(BufferType, netapi32defs.NetSetupDomainName.to_bytes(4, 'little'))

        return netapi32defs.NERR_Success



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import math
import struct

import speakeasy.winenv.arch as e_arch
import speakeasy.winenv.defs.windows.windows as windef

from .. import api

EINVAL = 22
_TRUNCATE = 0xFFFFFFFF

TIME_BASE = 1576292568
RAND_BASE = 0
TICK_BASE = 86400000  # 1 day in millisecs


class Msvcrt(api.ApiHandler):
    """
    Implements functions from various versions of the C runtime on Windows
    """
    name = 'msvcrt'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Msvcrt, self).__init__(emu)

        self.stdin = 0
        self.stdout = 1
        self.stderr = 2

        self.rand_int = RAND_BASE

        self.funcs = {}
        self.data = {}
        self.wintypes = windef

        self.tick_counter = TICK_BASE

        super(Msvcrt, self).__get_hook_attrs__(self)

    def hex_to_double(self, x):
        x = x.to_bytes(8, 'little')
        x = struct.unpack('d', x)[0]
        return x

    def double_to_hex(self, x):
        return struct.unpack('<Q', struct.pack('<d', x))[0]

    @impdata('_acmdln')
    def _acmdln(self, ptr=0):
        """Command line global CRT variable"""

        cmdln = ptr
        _argv = self.emu.get_argv()
        _argv = " ".join(_argv).encode('utf-8')

        ptr_size = self.emu.get_ptr_size()

        if not ptr:
            cmdln = self.mem_alloc(len(_argv) + ptr_size,
                                   base=None, tag='api.msvcrt._acmdln')
            p_cmdln = cmdln + ptr_size
            self.emu.mem_write(cmdln, p_cmdln.to_bytes(ptr_size, 'little'))
            self.emu.mem_write(p_cmdln, _argv)
        return cmdln

    # Reference: https://wiki.osdev.org/Visual_C%2B%2B_Runtime
    @apihook('_initterm_e', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def _initterm_e(self, emu, argv, ctx={}):
        """
        static int _initterm_e(_PIFV * pfbegin,
                                 _PIFV * pfend)
        """

        pfbegin, pfend = argv

        rv = 0

        return rv

    @apihook('_initterm', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def _initterm(self, emu, argv, ctx={}):
        """static void _initterm (_PVFV * pfbegin, _PVFV * pfend)"""

        pfbegin, pfend = argv

        rv = 0

        return rv

    @apihook('__getmainargs', argc=5)
    def __getmainargs(self, emu, argv, ctx={}):
        """
        int __getmainargs(
            int * _Argc,
            char *** _Argv,
            char *** _Env,
            int _DoWildCard,
            _startupinfo * _StartInfo);
        """

        _Argc, _Argv, _Env, _DoWildCard, _StartInfo = argv
        rv = 0

        return rv

    @apihook('__wgetmainargs', argc=5)
    def __wgetmainargs(self, emu, argv, ctx={}):
        """
        int __wgetmainargs (
           int *_Argc,
           wchar_t ***_Argv,
           wchar_t ***_Env,
           int _DoWildCard,
           _startupinfo * _StartInfo);
        """

        _Argc, _Argv, _Env, _DoWildCard, _StartInfo = argv
        rv = 0

        return rv

    @apihook('__p___wargv', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def __p___wargv(self, emu, argv, ctx={}):
        """WCHAR *** __p___wargv ()"""

        ptr_size = self.get_ptr_size()
        _argv = emu.get_argv()

        argv = [(a + '\x00\x00\x00\x00').encode('utf-16le') for a in _argv]
        array_size = (ptr_size * (len(argv) + 2))
        total = sum([len(a) for a in argv])
        total += array_size

        sptr = 0
        pptr = 0

        arg_mem = self.mem_alloc(size=total, tag='api.argv')
        pptr = arg_mem + ptr_size
        self.mem_write(arg_mem, pptr.to_bytes(ptr_size, 'little'))
        sptr = pptr + array_size

        for a in argv:
            self.mem_write(pptr, sptr.to_bytes(ptr_size, 'little'))
            pptr += ptr_size
            self.mem_write(sptr, a)
            sptr += len(a)
        self.mem_write(pptr, b'\x00' * ptr_size)
        rv = arg_mem

        # TODO: dispatch the VFV function array
        return rv

    @apihook('__p___argv', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def __p___argv(self, emu, argv, ctx={}):
        """char *** __p___argv ()"""

        ptr_size = self.get_ptr_size()
        _argv = emu.get_argv()

        argv = [(a + '\x00\x00\x00\x00').encode('utf-8') for a in _argv]

        array_size = (ptr_size * (len(argv) + 1))
        total = sum([len(a) for a in argv])
        total += array_size

        sptr = 0
        pptr = 0

        arg_mem = self.mem_alloc(size=total, tag='api.argv')
        pptr = arg_mem + ptr_size
        self.mem_write(arg_mem, pptr.to_bytes(ptr_size, 'little'))
        sptr = pptr + array_size

        for a in argv:
            self.mem_write(pptr, sptr.to_bytes(ptr_size, 'little'))
            pptr += ptr_size
            self.mem_write(sptr, a)
            sptr += len(a)
        self.mem_write(pptr, b'\x00' * ptr_size)

        rv = arg_mem
        return rv

    @apihook('__p___argc', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def __p___argc(self, emu, argv, ctx={}):
        """int * __p___argc ()"""

        _argv = emu.get_argv()

        argc = self.mem_alloc(size=4, tag='api.argc')
        self.mem_write(argc, len(_argv).to_bytes(4, 'little'))
        return argc

    @apihook('_get_initial_narrow_environment', argc=0,
             conv=e_arch.CALL_CONV_CDECL)
    def _get_initial_narrow_environment(self, emu, argv, ctx={}):
        """char** _get_initial_narrow_environment ()"""

        ptr_size = self.get_ptr_size()
        env = emu.get_env()
        total = ptr_size
        sptr = total
        pptr = 0
        fmt_env = []
        for k, v in env.items():
            envstr = '%s=%s\x00' % (k, v)
            envstr = envstr.encode('utf-8')
            total += len(envstr)
            fmt_env.append(envstr)
            total += ptr_size
            sptr += ptr_size

        envp = self.mem_alloc(size=total, tag='api.envp')
        pptr = envp
        sptr += envp

        for v in fmt_env:
            self.mem_write(pptr, sptr.to_bytes(ptr_size, 'little'))
            pptr += ptr_size
            self.mem_write(sptr, v)
            sptr += len(v)

        return envp

    @apihook('_get_initial_wide_environment', argc=0,
             conv=e_arch.CALL_CONV_CDECL)
    def _get_initial_wide_environment(self, emu, argv, ctx={}):
        """WCHAR** _get_initial_wide_environment ()"""

        ptr_size = self.get_ptr_size()
        env = emu.get_env()
        total = ptr_size
        sptr = total
        pptr = 0
        fmt_env = []
        for k, v in env.items():
            envstr = '%s=%s\x00' % (k, v)
            envstr = envstr.encode('utf-16le')
            total += len(envstr)
            fmt_env.append(envstr)
            total += ptr_size
            sptr += ptr_size

        envp = self.mem_alloc(size=total, tag='api.envp')
        pptr = envp
        sptr += envp

        for v in fmt_env:
            self.mem_write(pptr, sptr.to_bytes(ptr_size, 'little'))
            pptr += ptr_size
            self.mem_write(sptr, v)
            sptr += len(v)

        return envp

    @apihook('exit', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def exit(self, emu, argv, ctx={}):
        """
        void exit(
           int const status
        );
        """

        self.exit_process()

    @apihook('_exit', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _exit(self, emu, argv, ctx={}):
        """
        void _exit(
           int const status
        );
        """

        self.exit_process()

    @apihook('__acrt_iob_func', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def __acrt_iob_func(self, emu, argv, ctx={}):
        """FILE * __acrt_iob_func (fd)"""

        fd, = argv

        return fd

    @apihook('pow', argc=2, conv=e_arch.CALL_CONV_FLOAT)
    def pow(self, emu, argv, ctx={}):
        """
        double pow(
           double x,
           double y
        );
        """
        x, y = argv

        x = self.hex_to_double(x)
        y = self.hex_to_double(y)

        z = pow(x, y)

        z = self.double_to_hex(z)

        return z

    @apihook('floor', argc=1, conv=e_arch.CALL_CONV_FLOAT)
    def floor(self, emu, argv, ctx={}):
        """
        double floor(
           double x
        );
        """
        x, = argv

        y = self.hex_to_double(x)
        z = math.floor(y)
        z = self.double_to_hex(z)

        return z

    @apihook('sin', argc=1, conv=e_arch.CALL_CONV_FLOAT)
    def sin(self, emu, argv, ctx={}):
        """
        double sin(
           double x
        );
        """
        x, = argv

        y = self.hex_to_double(x)
        z = math.sin(y)
        z = self.double_to_hex(z)

        return z

    @apihook('abs', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def abs(self, emu, argv, ctx={}):
        """
        int abs(
           int x
        );
        """
        x, = argv
        y = abs(x)
        return y

    @apihook('strstr', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def strstr(self, emu, argv, ctx={}):
        """
        char *strstr(
           const char *str,
           const char *strSearch
        );
        """
        hay, needle = argv

        if hay:
            _hay = self.read_mem_string(hay, 1)
            argv[0] = _hay

        if needle:
            needle = self.read_mem_string(needle, 1)
            argv[1] = needle

        ret = _hay.find(needle)
        if ret != -1:
            ret = hay + ret
        else:
            ret = 0

        return ret

    @apihook('wcsstr', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def wcsstr(self, emu, argv, ctx={}):
        """
        wchar_t *wcsstr(
            const wchar_t *str,
            const wchar_t *strSearch
        );
        """
        hay, needle = argv

        if hay:
            _hay = self.read_mem_string(hay, 2)
            argv[0] = _hay

        if needle:
            needle = self.read_mem_string(needle, 2)
            argv[1] = needle

        ret = _hay.find(needle)
        if ret != -1:
            ret = hay + ret
        else:
            ret = 0

        return ret

    @apihook('strncat_s', argc=4, conv=e_arch.CALL_CONV_CDECL)
    def strncat_s(self, emu, argv, ctx={}):
        """
        errno_t strncat_s(
           char *strDest,
           size_t numberOfElements,
           const char *strSource,
           size_t count
        );
        """
        strDest, num, src, count = argv
        rv = 0

        is_truncated = (0xFFFFFFFF & count)
        if is_truncated == _TRUNCATE:
            is_truncated = True
        else:
            is_truncated = False

        argv[0] = self.read_mem_string(strDest, 1)
        argv[2] = self.read_mem_string(src, 1)

        slen1 = self.mem_string_len(strDest, 1)
        rem = num - slen1

        if is_truncated:
            if rem < count:
                self.mem_copy(strDest + slen1, src, count-1)
            else:
                self.mem_copy(strDest + slen1, src, count)
        else:
            if rem < count:
                rv = EINVAL
            else:
                self.mem_copy(strDest + slen1, src, count)

        return rv

    @apihook('__stdio_common_vfprintf', argc=e_arch.VAR_ARGS,
             conv=e_arch.CALL_CONV_CDECL)
    def __stdio_common_vfprintf(self, emu, argv, ctx={}):

        arch = emu.get_arch()
        if arch == e_arch.ARCH_AMD64:
            opts, stream, fmt, _, va_list = \
                emu.get_func_argv(e_arch.CALL_CONV_CDECL, 5)[:5]
        else:
            opts, opts2, stream, fmt, _, va_list = \
                emu.get_func_argv(e_arch.CALL_CONV_CDECL, 6)[:6]

        rv = 0

        fmt_str = self.read_mem_string(fmt, 1)
        fmt_cnt = self.get_va_arg_count(fmt_str)

        vargs = self.va_args(va_list, fmt_cnt)
        fin = self.do_str_format(fmt_str, vargs)

        argv[:] = [opts, stream, fin]

        rv = len(fin)
        return rv

    @apihook('memset', argc=3, conv=e_arch.CALL_CONV_CDECL)
    def memset(self, emu, argv, ctx={}):
        """
        void *memset ( void * ptr,
                       int value,
                       size_t num );
        """

        ptr, value, num = argv

        data = value.to_bytes(1, 'little') * num
        self.mem_write(ptr, data)

        return ptr

    @apihook('time', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def time(self, emu, argv, ctx={}):
        """
        time_t time( time_t *destTime );
        """

        destTime, = argv

        out_time = TIME_BASE
        if destTime:
            self.mem_write(destTime, out_time)

        return out_time

    @apihook('clock', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def clock(self, emu, argv, ctx={}):
        '''
        clock_t clock( void );
        '''

        self.tick_counter += 200

        return self.tick_counter

    @apihook('srand', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def srand(self, emu, argv, ctx={}):
        """
        void srand (unsigned int seed);
        """

        seed, = argv

        return

    @apihook('sprintf', argc=e_arch.VAR_ARGS, conv=e_arch.CALL_CONV_CDECL)
    def sprintf(self, emu, argv, ctx={}):
        """
        int sprintf(
            char *buffer,
            const char *format [,
            argument] ...
            );
        """
        buf, fmt = emu.get_func_argv(e_arch.CALL_CONV_CDECL, 2)
        fmt_str = self.read_string(fmt)
        fmt_cnt = self.get_va_arg_count(fmt_str)
        if not fmt_cnt:
            self.write_string(fmt_str, buf)
            return len(fmt_str)

        _argv = emu.get_func_argv(e_arch.CALL_CONV_CDECL, 2 + fmt_cnt)[2:]
        fin = self.do_str_format(fmt_str, _argv)

        self.write_string(fin, buf)
        argv.clear()
        argv.append(fin)
        return len(fin)

    @apihook('_snprintf', argc=e_arch.VAR_ARGS, conv=e_arch.CALL_CONV_CDECL)
    def _snprintf(self, emu, argv, ctx={}):
        """
        int _snprintf(
        char *buffer,
        size_t count,
        const char *format [,
        argument] ...
        );
        """
        buf, count, fmt = emu.get_func_argv(e_arch.CALL_CONV_CDECL, 3)
        fmt_str = self.read_string(fmt)
        fmt_cnt = self.get_va_arg_count(fmt_str)
        if not fmt_cnt:
            self.write_string(fmt_str, buf)
            return len(fmt_str)

        _argv = emu.get_func_argv(e_arch.CALL_CONV_CDECL, 3 + fmt_cnt)[3:]
        fin = self.do_str_format(fmt_str, _argv)

        self.write_string(fin, buf)
        argv.clear()
        argv.append(fin)
        return len(fin)

    @apihook('atoi', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def atoi(self, emu, argv, ctx={}):
        """
        int atoi(
            const char *str
        );
        """

        _str, = argv

        i = self.read_string(_str)
        argv[0] = i

        try:
            rv = int(i)
        except ValueError:
            rv = 0

        return rv

    @apihook('rand', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def rand(self, emu, argv, ctx={}):
        """
        int rand( void );
        """

        self.rand_int += 1

        return self.rand_int

    @apihook('__set_app_type', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def __set_app_type(self, emu, argv, ctx={}):
        """
        void __set_app_type (
            int at
        )
        """
        return

    @apihook('_set_app_type', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _set_app_type(self, emu, argv, ctx={}):
        return

    @apihook('__p__fmode', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def __p__fmode(self, emu, argv, ctx={}):
        """
        int* __p__fmode();
        """
        _O_TEXT = 0x4000

        ptr = self.mem_alloc(4, tag='api.fmode')
        data = _O_TEXT.to_bytes(4, 'little')
        self.mem_write(ptr, data)
        return ptr

    @apihook('__p__commode', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def __p__commode(self, emu, argv, ctx={}):
        """
        int* __p__commode();
        """
        _IOCOMMIT = 0x4000

        ptr = self.mem_alloc(4, tag='api.commode')
        data = _IOCOMMIT.to_bytes(4, 'little')
        self.mem_write(ptr, data)
        return ptr

    @apihook('_controlfp', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def _controlfp(self, emu, argv, ctx={}):
        """
        unsigned int _controlfp(unsigned int new,
                                unsinged int mask)
        """
        return 0

    @apihook('strcpy', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def strcpy(self, emu, argv, ctx={}):
        """
        char *strcpy(
           char *strDestination,
           const char *strSource
        );
        """
        dest, src = argv
        s = self.read_string(src)

        self.write_string(s, dest)
        argv[1] = s
        return dest

    @apihook('wcscpy', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def wcscpy(self, emu, argv, ctx={}):
        """
        wchar_t *wcscpy(
            wchar_t *strDestination,
            const wchar_t *strSource
        );
        """
        dest, src = argv
        ws = self.read_wide_string(src)
        self.write_wide_string(ws, dest)
        argv[1] = ws
        return dest

    @apihook('strncpy', argc=3, conv=e_arch.CALL_CONV_CDECL)
    def strncpy(self, emu, argv, ctx={}):
        """
        char * strncpy(
            char * destination,
            const char * source,
            size_t num
        );
        """
        dest, src, length = argv
        s = self.read_string(src, max_chars=length)
        if len(s) < length:
            s += '\x00'*(length-len(s))
        self.write_string(s, dest)
        argv[1] = s
        return dest

    @apihook('memcpy', argc=3, conv=e_arch.CALL_CONV_CDECL)
    def memcpy(self, emu, argv, ctx={}):
        """
        void *memcpy(
            void *dest,
            const void *src,
            size_t count
            );
        """
        dest, src, count = argv
        data = self.mem_read(src, count)
        self.mem_write(dest, data)
        return dest

    @apihook('memmove', argc=3, conv=e_arch.CALL_CONV_CDECL)
    def memmove(self, emu, argv, ctx={}):
        """
        void *memmove(
            void *dest,
            const void *src,
            size_t count
        );
        """
        dest, src, count = argv
        data = self.mem_read(src, count)
        self.mem_write(dest, data)
        return dest

    @apihook('memcmp', argc=3, conv=e_arch.CALL_CONV_CDECL)
    def memcmp(self, emu, argv, ctx={}):
        """
        int memcmp(
           const void *buffer1,
           const void *buffer2,
           size_t count
        );
        """
        diff = 0
        buff1, buff2, cnt = argv
        for i in range(cnt):
            b1 = self.mem_read(buff1, 1)
            b2 = self.mem_read(buff2, 1)
            if b1 > b2:
                diff = 1
                break
            elif b1 < b2:
                diff = -1
                break

        return diff

    @apihook('_except_handler4_common', argc=6, conv=e_arch.CALL_CONV_CDECL)
    def _except_handler4_common(self, emu, argv, ctx={}):
        """
        _CRTIMP  __C_specific_handler(
        _In_    struct _EXCEPTION_RECORD   *ExceptionRecord,
        _In_    void                       *EstablisherFrame,
        _Inout_ struct _CONTEXT            *ContextRecord,
        _Inout_ struct _DISPATCHER_CONTEXT *DispatcherContext
        );
        """
        # Inferred from the SEH teardowns described here:
        # https://bytepointer.com/resources/pietrek_crash_course_depths_of_win32_seh.htm
        # http://www.openrce.org/articles/full_view/21

        # Two additional arguments are pushed to the function to check security cookies
        cookie_ptr, cookie_func, record, frame, context, dispath_ctx = argv
        rv = 0

        cookie = self.mem_read(cookie_ptr, 4)
        cookie = int.from_bytes(cookie, 'little')

        thread = emu.get_current_thread()

        # Break down the exception records into something more manageable
        curr_frame = frame
        seh = thread.get_seh()

        _ctx = self.wintypes.CONTEXT(emu.get_ptr_size())
        _ctx = self.mem_cast(_ctx, context)

        seh.set_context(_ctx, address=context)
        seh.set_record(record)

        seh.clear_frames()

        while curr_frame != 0:
            reg = self.wintypes.EXCEPTION_REGISTRATION(emu.get_ptr_size())
            reg = self.mem_cast(reg, curr_frame)

            scope_table = reg.ScopeTable ^ cookie

            st = self.wintypes.EH4_SCOPETABLE(emu.get_ptr_size())
            st = self.mem_cast(st, scope_table)

            rec = self.wintypes.EH4_SCOPETABLE_RECORD(emu.get_ptr_size())
            # The trylevel will tell us what scope record to get
            scope_record_offset = scope_table + st.sizeof()
            tl = reg.TryLevel
            if reg.TryLevel & 0x80000000:
                tl = -0x100000000 + reg.TryLevel

            if tl == -2:  # -2 is the outermost scope
                tl = 0

            scope_record_offset += (rec.sizeof() * tl)
            rec = self.mem_cast(rec, scope_record_offset)

            seh.add_frame(reg, st, [rec, ])

            curr_frame = reg.Next

        return rv

    @apihook('_seh_filter_exe', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def _seh_filter_exe(self, emu, argv, ctx={}):
        """
        int __cdecl _seh_filter_exe(
           unsigned long _ExceptionNum,
           struct _EXCEPTION_POINTERS* _ExceptionPtr
        );
        """
        except_num, exc_ptr = argv
        rv = 1

        return rv

    @apihook('_except_handler3', argc=4, conv=e_arch.CALL_CONV_CDECL)
    def _except_handler3(self, emu, argv, ctx={}):
        """
        int _except_handler3(
        PEXCEPTION_RECORD exception_record,
        PEXCEPTION_REGISTRATION registration,
        PCONTEXT context,
        PEXCEPTION_REGISTRATION dispatcher
        );
        """
        rv = 1
        return rv

    @apihook('_seh_filter_dll', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def _seh_filter_dll(self, emu, argv, ctx={}):
        """
        int __cdecl _seh_filter_dll(
           unsigned long _ExceptionNum,
           struct _EXCEPTION_POINTERS* _ExceptionPtr
        );
        """
        except_num, exc_ptr = argv
        rv = 1

        return rv

    @apihook('puts', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def puts(self, emu, argv, ctx={}):
        """
        int puts(
           const char *str
        );
        """
        s, = argv

        string = self.read_mem_string(s, 1)
        argv[0] = string
        rv = len(string)

        return rv

    @apihook('_initialize_onexit_table', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _initialize_onexit_table(self, emu, argv, ctx={}):
        """
        int _initialize_onexit_table(
            _onexit_table_t* table
            );
        """
        rv = 0

        return rv

    @apihook('_register_onexit_function', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def _register_onexit_function(self, emu, argv, ctx={}):
        """
        int _register_onexit_function(
            _onexit_table_t* table,
            _onexit_t        function
            );
        """
        rv = 0

        return rv

    @apihook('malloc', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def malloc(self, emu, argv, ctx={}):
        """
        void *malloc(
        size_t size
        );
        """
        size, = argv

        chunk = self.heap_alloc(size, heap='HeapAlloc')
        return chunk

    @apihook('calloc', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def calloc(self, emu, argv, ctx={}):
        """
        void *calloc(
        size_t num,
        size_t size
        );
        """
        num, size, = argv

        chunk = self.heap_alloc(num*size, heap='HeapAlloc')

        buf = b'\x00' * (num*size)
        self.mem_write(chunk, buf)

        return chunk

    @apihook('free', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def free(self, emu, argv, ctx={}):
        """
        void free(
        void *memblock
        );
        """
        mem, = argv
        self.mem_free(mem)

    @apihook('_beginthreadex', argc=6, conv=e_arch.CALL_CONV_CDECL)
    def _beginthreadex(self, emu, argv, ctx={}):
        """
        uintptr_t _beginthreadex(
            void *security,
            unsigned stack_size,
            unsigned ( __stdcall *start_address )( void * ),
            void *arglist,
            unsigned initflag,
            unsigned *thrdaddr
        );
        """
        security, stack_size, start_address, arglist, initflag, thrdaddr = argv

        handle, obj = self.create_thread(start_address, arglist, emu.get_current_process())

        if thrdaddr:
            self.mem_write(thrdaddr, obj.get_id().to_bytes(4, 'little'))

        return handle

    @apihook('_beginthread', argc=3, conv=e_arch.CALL_CONV_CDECL)
    def _beginthread(self, emu, argv, ctx={}):
        """
        uintptr_t _beginthread
        void( __cdecl *start_address )( void * ),
        unsigned stack_size,
        void *arglist
        );
        """
        start_address, stack_size, arglist = argv

        handle, obj = self.create_thread(start_address, arglist, emu.get_current_process())
        return handle

    @apihook('system', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def system(self, emu, argv, ctx={}):
        """
        int system(
           const char *command
        );
        """
        s, = argv

        string = self.read_mem_string(s, 1)
        argv[0] = string
        rv = len(string)

        return rv

    @apihook('toupper', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def toupper(self, emu, argv, ctx={}):
        """
        int toupper(
           int c
        );
        """
        c, = argv
        argv[0] = c
        if 0x00 <= c <= 0x7f:
            c = ord(chr(c).upper())
        else:
            c = 0x00
        return c

    @apihook('strlen', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def strlen(self, emu, argv, ctx={}):
        """
        size_t strlen(
            const char *str
        );
        """
        s, = argv

        string = self.read_mem_string(s, 1)
        argv[0] = string
        rv = len(string)

        return rv

    @apihook('strcat', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def strcat(self, emu, argv, ctx={}):
        '''
        char *strcat(
            char *strDestination,
            const char *strSource
        );
        '''
        _str1, _str2 = argv
        s1 = self.read_mem_string(_str1, 1)
        s2 = self.read_mem_string(_str2, 1)
        argv[0] = s1
        argv[1] = s2
        new = (s1 + s2).encode('utf-8')
        self.mem_write(_str1, new + b'\x00')
        return _str1

    @apihook('wcscat', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def wcscat(self, emu, argv, ctx={}):
        '''
        wchar_t *wcscat(
           wchar_t *strDestination,
           const wchar_t *strSource
        );
        '''
        _str1, _str2 = argv
        s1 = self.read_mem_string(_str1, 2)
        s2 = self.read_mem_string(_str2, 2)
        argv[0] = s1
        argv[1] = s2
        new = (s1 + s2).encode('utf-16le')
        self.mem_write(_str1, new + b'\x00\x00')
        return _str1

    @apihook('wcslen', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def wcslen(self, emu, argv, ctx={}):
        """
        size_t wcslen(
          const wchar_t* wcs
        );
        """
        s, = argv
        string = self.read_wide_string(s)
        argv[0] = string
        rv = len(string)

        return rv

    @apihook('_lock', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _lock(self, emu, argv, ctx={}):
        """
        void __cdecl _lock
            int locknum
        );
        """
        return

    @apihook('_unlock', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _unlock(self, emu, argv, ctx={}):
        """
        void __cdecl _unlock
            int locknum
        );
        """
        return

    @apihook('_ltoa', argc=3, conv=e_arch.CALL_CONV_CDECL)
    def _ltoa(self, emu, argv, ctx={}):
        """
        char *_ltoa(
            long value,
            char *str,
            int radix
        );
        """
        val, out_str, radix, = argv

        v = str(val).encode('utf-8')
        self.mem_write(out_str, v)
        return

    @apihook('__dllonexit', argc=3, conv=e_arch.CALL_CONV_CDECL)
    def __dllonexit(self, emu, argv, ctx={}):
        """
        onexit_t __dllonexit(
            _onexit_t func,
            _PVFV **  pbegin,
            _PVFV **  pend
        )
        """
        func, pbegin, pend, = argv
        return func

    @apihook('strncmp', argc=3, conv=e_arch.CALL_CONV_CDECL)
    def strncmp(self, emu, argv, ctx={}):
        """
        int strncmp(
            const char *string1,
            const char *string2,
            size_t count
        );
        """
        s1, s2, c = argv
        rv = 1

        string1 = self.read_mem_string(s1, 1)
        string2 = self.read_mem_string(s2, 1)
        if string1 == string2:
            rv = 0
        argv[0] = string1
        argv[1] = string2

        return rv

    @apihook('strcmp', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def strcmp(self, emu, argv, ctx={}):
        """
        int strcmp(
            const char *string1,
            const char *string2,
        );
        """
        s1, s2 = argv
        rv = 1

        string1 = self.read_mem_string(s1, 1)
        string2 = self.read_mem_string(s2, 1)
        if string1 == string2:
            rv = 0
        argv[0] = string1
        argv[1] = string2

        return rv

    @apihook('strrchr', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def strrchr(self, emu, argv, ctx={}):
        """
        char *strrchr(
            const char *str,
            int c
            );
        """
        cstr, c = argv
        cs = self.read_string(cstr)
        hay = cs.encode('utf-8')
        needle = c.to_bytes(1, 'little')

        offset = hay.rfind(needle)
        if offset < 0:
            rv = 0
        else:
            rv = cstr + offset

        argv[0] = cs
        argv[1] = needle.decode('utf-8')

        return rv

    @apihook('_ftol', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _ftol(self, emu, argv, ctx={}):
        """
        int _ftol(int);
        """
        f, = argv
        return int(f)

    @apihook('_adjust_fdiv', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def _adjust_fdiv(self, emu, argv, ctx={}):
        """
        void _adjust_fdiv(void)
        """
        return

    @apihook('tolower', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def tolower(self, emu, argv, ctx={}):
        """
        int tolower ( int c );
        """
        c, = argv
        return c | 0x20

    @apihook('sscanf', argc=e_arch.VAR_ARGS, conv=e_arch.CALL_CONV_CDECL)
    def sscanf(self, emu, argv, ctx={}):
        """
        int sscanf ( const char * s, const char * format, ...);
        """
        return

    @apihook('strchr', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def strchr(self, emu, argv, ctx={}):
        """
        char *strchr(
            const char *str,
            int c
            );
        """
        cstr, c = argv
        cs = self.read_string(cstr)
        hay = cs.encode('utf-8')
        needle = c.to_bytes(1, 'little')

        offset = hay.find(needle)
        if offset < 0:
            rv = 0
        else:
            rv = cstr + offset

        argv[0] = cs
        argv[1] = needle.decode('utf-8')

        return rv

    @apihook('_set_invalid_parameter_handler', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _set_invalid_parameter_handler(self, emu, argv, ctx={}):
        """
        _invalid_parameter_handler _set_invalid_parameter_handler(
        _invalid_parameter_handler pNew
        );
        """
        pNew, = argv

        return 0

    @apihook('__CxxFrameHandler', argc=4, conv=e_arch.CALL_CONV_CDECL)
    def __CxxFrameHandler(self, emu, argv, ctx={}):
        """
        EXCEPTION_DISPOSITION __CxxFrameHandler(
            EHExceptionRecord  *pExcept,
            EHRegistrationNode *pRN,
            void               *pContext,
            DispatcherContext  *pDC
        )
        """
        pExcept, pRN, pContext, pDC, = argv
        return 0

    @apihook('_vsnprintf', argc=4, conv=e_arch.CALL_CONV_CDECL)
    def _vsnprintf(self, emu, argv, ctx={}):
        """
        int _vsnprintf(
            char *buffer,
            size_t count,
            const char *format,
            va_list argptr
        );
        """
        buffer, count, _format, argptr = argv
        rv = 0

        fmt_str = self.read_mem_string(_format, 1)
        fmt_cnt = self.get_va_arg_count(fmt_str)

        vargs = self.va_args(argptr, fmt_cnt)

        fin = self.do_str_format(fmt_str, vargs)
        fin = fin[:count] + '\x00'

        rv = len(fin)
        self.mem_write(buffer, fin.encode('utf-8'))
        argv[0] = fin.replace('\x00', '')
        argv[1] = fmt_str

        return rv

    @apihook('__stdio_common_vsprintf', argc=7, conv=e_arch.CALL_CONV_CDECL)
    def __stdio_common_vsprintf(self, emu, argv, ctx={}):
        """
        int __stdio_common_vsprintf(
            unsigned int64 Options,
            char *Buffer,
            unsigned int BufferCount,
            const char *format,
            locale_t Locale,
            va_list argptr
        );
        """
        options_lo, options_hi, buffer, count, _format, locale, argptr = argv
        rv = 0
        fmt_str = self.read_mem_string(_format, 1)
        fmt_cnt = self.get_va_arg_count(fmt_str)

        vargs = self.va_args(argptr, fmt_cnt)

        fin = self.do_str_format(fmt_str, vargs)
        fin = fin[:count] + '\x00'

        rv = len(fin)
        self.mem_write(buffer, fin.encode('utf-8'))
        argv[0] = fin.replace('\x00', '')
        argv[1] = fmt_str

        return rv

    @apihook('_strcmpi', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def _strcmpi(self, emu, argv, ctx={}):
        """
        int _strcmpi(
                const char *string1,
                const char *string2
                );
        """
        string1, string2 = argv
        rv = 1

        if not string1 or not string2:
            return rv

        cs1 = self.read_string(string1)
        cs2 = self.read_string(string2)

        argv[0] = cs1
        argv[1] = cs2

        if cs1.lower() == cs2.lower():
            rv = 0

        return rv

    @apihook('_wcsicmp', argc=2, conv=e_arch.CALL_CONV_CDECL)
    def _wcsicmp(self, emu, argv, ctx={}):
        """
        int _wcsicmp(
                const wchar_t *string1,
                const wchar_t *string2
                );
        """
        string1, string2 = argv
        rv = 1

        if not string1 or not string2:
            return rv

        cs1 = self.read_wide_string(string1)
        cs2 = self.read_wide_string(string2)

        argv[0] = cs1
        argv[1] = cs2

        if cs1.lower() == cs2.lower():
            rv = 0

        return rv

    @apihook('??3@YAXPAX@Z', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def __3_YAXPAX_Z(self, emu, argv, ctx={}):
        return

    @apihook('??2@YAPAXI@Z', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def __2_YAPAXI_Z(self, emu, argv, ctx={}):
        return

    @apihook('__current_exception_context', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def __current_exception_context(self, emu, argv, ctx={}):
        return

    @apihook('__current_exception', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def __current_exception(self, emu, argv, ctx={}):
        return

    @apihook('_set_new_mode', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _set_new_mode(self, emu, argv, ctx={}):
        return

    @apihook('_configthreadlocale', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _configthreadlocale(self, emu, argv, ctx={}):
        return

    @apihook('_setusermatherr', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _setusermatherr(self, emu, argv, ctx={}):
        return

    @apihook('__setusermatherr', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def __setusermatherr(self, emu, argv, ctx={}):
        return

    @apihook('_cexit', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def _cexit(self, emu, argv, ctx={}):
        # TODO: handle atexit flavor functions
        self.exit_process()

    @apihook('_c_exit', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def _c_exit(self, emu, argv, ctx={}):
        self.exit_process()

    @apihook('_register_thread_local_exe_atexit_callback', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _register_thread_local_exe_atexit_callback(self, emu, argv, ctx={}):
        return

    @apihook('_crt_atexit', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _crt_atexit(self, emu, argv, ctx={}):
        return

    @apihook('_controlfp_s', argc=3, conv=e_arch.CALL_CONV_CDECL)
    def _controlfp_s(self, emu, argv, ctx={}):
        return

    @apihook('terminate', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def terminate(self, emu, argv, ctx={}):
        self.exit_process()

    @apihook('_crt_atexit', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _crt_atexit(self, emu, argv, ctx={}):
        return

    @apihook('_initialize_narrow_environment', argc=0, conv=e_arch.CALL_CONV_CDECL)
    def _initialize_narrow_environment(self, emu, argv, ctx={}):
        return

    @apihook('_configure_narrow_argv', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _configure_narrow_argv(self, emu, argv, ctx={}):
        return

    @apihook('_set_fmode', argc=1, conv=e_arch.CALL_CONV_CDECL)
    def _set_fmode(self, emu, argv, ctx={}):
        return

    @apihook('_itoa', argc=3, conv=e_arch.CALL_CONV_CDECL)
    def _itoa(self, emu, argv, ctx={}):
        return

    @apihook('_itow', argc=3, conv=e_arch.CALL_CONV_CDECL)
    def _itow(self, emu, argv, ctx={}):
        return



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import uuid

from speakeasy.struct import EmuStruct, Ptr

S_OK = 0

RPC_C_AUTHN_LEVEL_DEFAULT = 0
RPC_C_AUTHN_LEVEL_NONE = 1
RPC_C_AUTHN_LEVEL_CONNECT = 2
RPC_C_AUTHN_LEVEL_CALL = 3
RPC_C_AUTHN_LEVEL_PKT = 4
RPC_C_AUTHN_LEVEL_PKT_INTEGRITY = 5
RPC_C_AUTHN_LEVEL_PKT_PRIVACY = 6

RPC_C_IMP_LEVEL_DEFAULT = 0
RPC_C_IMP_LEVEL_ANONYMOUS = 1
RPC_C_IMP_LEVEL_IDENTIFY = 2
RPC_C_IMP_LEVEL_IMPERSONATE = 3
RPC_C_IMP_LEVEL_DELEGATE = 4

CLSID_WbemLocator = '{4590F811-1D3A-11D0-891F-00AA004B2E24}'
CLSID_IWbemContext = '{674B6698-EE92-11D0-AD71-00C04FD8FDFF}'

IID_IWbemLocator = '{DC12A687-737F-11CF-884D-00AA004B2E24}'
IID_IWbemContext = '{44ACA674-E8FC-11D0-A07C-00C04FB68820}'


class ComInterface(object):
    def __init__(self, iface, name, ptr_size):
        self.iface = iface(ptr_size)
        self.address = 0
        self.name = name


class IUnknown(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.QueryInterface = Ptr
        self.AddRef = Ptr
        self.Release = Ptr


class IMalloc(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.IUnknown = IUnknown
        self.Alloc = Ptr
        self.Realloc = Ptr
        self.Free = Ptr
        self.GetSize = Ptr
        self.DidAlloc = Ptr
        self.HeapMinimize = Ptr


class IWbemLocator(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.IUnknown = IUnknown
        self.ConnectServer = Ptr


class IWbemServices(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.IUnknown = IUnknown
        self.OpenNamespace = Ptr
        self.CancelAsyncCall = Ptr
        self.QueryObjectSink = Ptr
        self.GetObject = Ptr
        self.GetObjectAsync = Ptr
        self.PutClass = Ptr
        self.PutClassAsync = Ptr
        self.DeleteClass = Ptr
        self.DeleteClassAsync = Ptr
        self.CreateClassEnum = Ptr
        self.CreateClassEnumAsync = Ptr
        self.PutInstance = Ptr
        self.PutInstanceAsync = Ptr
        self.DeleteInstance = Ptr
        self.DeleteInstanceAsync = Ptr
        self.CreateInstanceEnum = Ptr
        self.CreateInstanceEnumAsync = Ptr
        self.ExecQuery = Ptr
        self.ExecQueryAsync = Ptr
        self.ExecNotificationQuery = Ptr
        self.ExecNotificationQueryAsync = Ptr
        self.ExecMethod = Ptr
        self.ExecMethodAsync = Ptr


class IWbemContext(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.IUnknown = IUnknown
        self.Clone = Ptr
        self.GetNames = Ptr
        self.BeginEnumeration = Ptr
        self.Next = Ptr
        self.EndEnumeration = Ptr
        self.SetValue = Ptr
        self.GetValue = Ptr
        self.DeleteValue = Ptr
        self.DeleteAll = Ptr


IFACE_TYPES = {'IUnknown': IUnknown,
               'IMalloc':  IMalloc,
               'IWbemLocator': IWbemLocator,
               'IWbemServices': IWbemServices,
               'IWbemContext': IWbemContext}


def get_define_int(define, prefix=''):
    for k, v in globals().items():
        if not isinstance(v, int) or v != define:
            continue
        if prefix:
            if k.startswith(prefix):
                return k
        else:
            return k


def get_define_str(define, prefix=''):
    for k, v in globals().items():
        if not isinstance(v, str) or v != define:
            continue
        if prefix:
            if k.startswith(prefix):
                return k
        else:
            return k


def get_clsid(define):
    return get_define_str(define, prefix='CLSID_')


def get_iid(define):
    return get_define_str(define, prefix='IID_')


def get_rpc_authlevel(define):
    return get_define_int(define, prefix='RPC_C_AUTHN_LEVEL_')


def get_rcp_implevel(define):
    return get_define_int(define, prefix='RPC_C_IMP_LEVEL_')


def convert_guid_bytes_to_str(guid_bytes):
    u = uuid.UUID(bytes_le=guid_bytes)
    return ('{%s}' % u).upper()



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.


from .. import api


class Msimg32(api.ApiHandler):

    """
    Implements exported functions from msimg.dll
    """

    name = 'msimg32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Msimg32, self).__init__(emu)
        super(Msimg32, self).__get_hook_attrs__(self)

    @apihook('TransparentBlt', argc=11)
    def TransparentBlt(self, emu, argv, ctx={}):
        '''
        BOOL TransparentBlt(
          HDC  hdcDest,
          int  xoriginDest,
          int  yoriginDest,
          int  wDest,
          int  hDest,
          HDC  hdcSrc,
          int  xoriginSrc,
          int  yoriginSrc,
          int  wSrc,
          int  hSrc,
          UINT crTransparent
        );
        '''
        return 1





CSIDL = {
    0x00: "CSIDL_DESKTOP",
    0x01: "CSIDL_INTERNET",
    0x02: "CSIDL_PROGRAMS",
    0x03: "CSIDL_CONTROLS",
    0x04: "CSIDL_PRINTERS",
    0x05: "CSIDL_MYDOCUMENTS",
    0x06: "CSIDL_FAVORITES",
    0x07: "CSIDL_STARTUP",
    0x08: "CSIDL_RECENT",
    0x09: "CSIDL_SENDTO",
    0x0a: "CSIDL_BITBUCKET",
    0x0b: "CSIDL_STARTMENU",
    0x0d: "CSIDL_MYMUSIC",
    0x0e: "CSIDL_MYVIDEO",
    0x10: "CSIDL_DESKTOPDIRECTORY",
    0x11: "CSIDL_DRIVES",
    0x12: "CSIDL_NETWORK",
    0x13: "CSIDL_NETHOOD",
    0x14: "CSIDL_FONTS",
    0x15: "CSIDL_TEMPLATES",
    0x16: "CSIDL_COMMON_STARTMENU",
    0x17: "CSIDL_COMMON_PROGRAMS",
    0x18: "CSIDL_COMMON_STARTUP",
    0x19: "CSIDL_COMMON_DESKTOPDIRECTORY",
    0x1a: "CSIDL_APPDATA",
    0x1b: "CSIDL_PRINTHOOD",
    0x1c: "CSIDL_LOCAL_APPDATA",
    0x1d: "CSIDL_ALTSTARTUP",
    0x1e: "CSIDL_COMMON_ALTSTARTUP",
    0x1f: "CSIDL_COMMON_FAVORITES",
    0x20: "CSIDL_INTERNET_CACHE",
    0x21: "CSIDL_COOKIES",
    0x22: "CSIDL_HISTORY",
    0x23: "CSIDL_COMMON_APPDATA",
    0x24: "CSIDL_WINDOWS",
    0x25: "CSIDL_SYSTEM",
    0x26: "CSIDL_PROGRAM_FILES",
    0x27: "CSIDL_MYPICTURES",
    0x28: "CSIDL_PROFILE",
    0x29: "CSIDL_SYSTEMX86",
    0x2a: "CSIDL_PROGRAM_FILESX86",
    0x2b: "CSIDL_PROGRAM_FILES_COMMON",
    0x2c: "CSIDL_PROGRAM_FILES_COMMONX86",
    0x2e: "CSIDL_COMMON_DOCUMENTS",
    0x2d: "CSIDL_COMMON_TEMPLATES",
    0x2f: "CSIDL_COMMON_ADMINTOOLS",
    0x30: "CSIDL_ADMINTOOLS",
    0x31: "CSIDL_CONNECTIONS",
    0x35: "CSIDL_COMMON_MUSIC",
    0x36: "CSIDL_COMMON_PICTURES",
    0x37: "CSIDL_COMMON_VIDEO",
    0x38: "CSIDL_RESOURCES",
    0x39: "CSIDL_RESOURCES_LOCALIZED",
    0x3b: "CSIDL_CDBURN_AREA",
    0x3d: "CSIDL_COMPUTERSNEARME",
    0x3f: "CSIDL_PLAYLISTS",
    0x40: "CSIDL_SAMPLE_MUSIC",
    0x41: "CSIDL_SAMPLE_PLAYLISTS",
    0x42: "CSIDL_SAMPLE_PICTURES",
    0x43: "CSIDL_SAMPLE_VIDEOS",
    0x45: "CSIDL_PHOTOALBUMS",
}



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import sys
import inspect

import speakeasy.winenv.arch as _arch
from speakeasy.errors import ApiEmuError
from speakeasy.winenv.api import api
from speakeasy.winenv.api.kernelmode import * # noqa
from speakeasy.winenv.api.usermode import * # noqa


def autoload_api_handlers():
    api_handlers = []

    for modname, modobj in sys.modules.items():
        if not modname.startswith(('speakeasy.winenv.api.kernelmode.',
                                   'speakeasy.winenv.api.usermode.')):
            continue
        for clsname, clsobj in inspect.getmembers(modobj, inspect.isclass):
            if clsobj is not api.ApiHandler and issubclass(clsobj, api.ApiHandler):
                api_handlers.append((clsobj.name, clsobj))

    return tuple(api_handlers)


API_HANDLERS = autoload_api_handlers()


class WindowsApi:

    def __init__(self, emu):
        self.mods = {}
        self.instances = []
        self.data = {}
        self.emu = emu
        arch = self.emu.get_arch()

        if arch == _arch.ARCH_X86:
            self.ptr_size = 4
        elif arch == _arch.ARCH_AMD64:
            self.ptr_size = 8
        else:
            raise ApiEmuError('Invalid architecture')

    def load_api_handler(self, mod_name):
        for name, hdl in API_HANDLERS:
            name = name.lower()
            if mod_name and name == mod_name.lower():
                handler = self.mods.get(name)
                if not handler:
                    handler = hdl(self.emu)
                    self.mods.update({name: handler})
                return handler
        return None

    def get_data_export_handler(self, mod_name, exp_name):
        mod = self.mods.get(mod_name)
        if not mod:
            mod = self.load_api_handler(mod_name)
        if not mod:
            return None, None
        return (mod, mod.get_data_handler(exp_name))

    def get_export_func_handler(self, mod_name, exp_name):
        mod = self.mods.get(mod_name)
        if not mod:
            mod = self.load_api_handler(mod_name)
        if not mod:
            return None, None
        return (mod, mod.get_func_handler(exp_name))

    def call_api_func(self, mod, func, argv, ctx):
        """
        Call the handler to implement the imported API
        """
        return func(mod, self.emu, argv, ctx)

    def call_data_func(self, mod, func, ptr):
        """
        Call the handler to initialize and return imported data variables
        """
        return func(mod, ptr)



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.arch as _arch

from .. import api


class Msi32(api.ApiHandler):

    name = 'msi32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Msi32, self).__init__(emu)
        super(Msi32, self).__get_hook_attrs__(self)

    @apihook('MsiDatabaseMergeA', argc=3, conv=_arch.CALL_CONV_STDCALL, ordinal=29)
    def MsiDatabaseMergeA(self, emu, argv, ctx={}):
        """
        UINT MsiDatabaseMergeA(
          MSIHANDLE hDatabase,
          MSIHANDLE hDatabaseMerge,
          LPCSTR    szTableName
        );
        """
        return 0



# Copyright (C) 2021 FireEye, Inc. All Rights Reserved.

from speakeasy.struct import EmuStruct, Ptr
import ctypes as ct

NERR_Success = 0

NetSetupUnknownStatus = 0
NetSetupUnjoined = 1
NetSetupWorkgroupName = 2
NetSetupDomainName = 3


class WKSTA_INFO_100(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.wki_platform_id = Ptr
        self.wki_computername = Ptr
        self.wki_langroup = Ptr
        self.wki_ver_major = ct.c_uint32
        self.wki_ver_minor = ct.c_uint32


class WKSTA_INFO_101(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.wki_platform_id = Ptr
        self.wki_computername = Ptr
        self.wki_langroup = Ptr
        self.wki_ver_major = ct.c_uint32
        self.wki_ver_minor = ct.c_uint32
        self.wki_lanroot = Ptr


class WKSTA_INFO_102(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.wki_platform_id = Ptr
        self.wki_computername = Ptr
        self.wki_langroup = Ptr
        self.wki_ver_major = ct.c_uint32
        self.wki_ver_minor = ct.c_uint32
        self.wki_lanroot = Ptr
        self.wki_logged_on_users = Ptr



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

__all__ = ["volmgr"] # noqa

from . import * # noqa


def _get_kmods():
    def imports():
        import types
        for name, val in globals().items():
            if isinstance(val, types.ModuleType):
                yield val
    kmods = []
    imps = list(imports())
    for i in imps:
        if 'DriverModule' in dir(i):
            kmods.append(i)
    return kmods



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import struct

from .. import api


class OleAut32(api.ApiHandler):

    name = 'oleaut32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(OleAut32, self).__init__(emu)
        super(OleAut32, self).__get_hook_attrs__(self)

    @apihook('SysAllocString', argc=1)
    def SysAllocString(self, emu, argv, ctx={}):
        """
        BSTR SysAllocString(
            const OLECHAR *psz
        );
        """
        psz, = argv
        alloc_str = self.read_mem_string(psz, 2)
        if alloc_str:
            argv[0] = alloc_str
            alloc_str += '\x00'
            ws = alloc_str.encode('utf-16le')
            ws_len = len(ws)

            # https://docs.microsoft.com/en-us/previous-versions/windows/desktop/automat/bstr
            bstr_len = 4 + ws_len
            bstr = self.mem_alloc(bstr_len)
            bstr_bytes = struct.pack('<I', ws_len - 2) + ws

            self.mem_write(bstr, bstr_bytes)

            return bstr + 4

        return 0

    @apihook('SysFreeString', argc=1)
    def SysFreeString(self, emu, argv, ctx={}):
        """
        void SysFreeString(
            BSTR bstrString
        );
        """
        argv[0] = self.read_wide_string(argv[0])
        return



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from .. import api
import speakeasy.winenv.defs.windows.com as comdefs


class ComApi(api.ApiHandler):
    """
    Implements COM interfaces
    """
    name = 'com_api'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(ComApi, self).__init__(emu)

        self.funcs = {}
        self.data = {}

        super(ComApi, self).__get_hook_attrs__(self)

    # First argument (self) is not reflected in method definitions; note this increases argc by 1
    @apihook('IUnknown.QueryInterface', argc=3)
    def IUnknown_QueryInterface(self, emu, argv, ctx={}):
        """
        HRESULT QueryInterface(
            REFIID riid,
            void   **ppvObject
        );
        """
        # not implemented
        return comdefs.S_OK

    @apihook('IUnknown.AddRef', argc=1)
    def IUnknown_AddRef(self, emu, argv, ctx={}):
        """
        ULONG AddRef();
        """
        # not implemented
        return 1

    @apihook('IUnknown.Release', argc=1)
    def IUnknown_Release(self, emu, argv, ctx={}):
        """
        ULONG Release();
        """
        # not implemented
        return 0

    @apihook('IWbemLocator.ConnectServer', argc=9)
    def IWbemLocator_ConnectServer(self, emu, argv, ctx={}):
        """
        HRESULT ConnectServer(
            const BSTR    strNetworkResource,
            const BSTR    strUser,
            const BSTR    strPassword,
            const BSTR    strLocale,
            long          lSecurityFlags,
            const BSTR    strAuthority,
            IWbemContext  *pCtx,
            IWbemServices **ppNamespace
        );
        """
        ptr, strNetworkResource, strUser, strPassword, strLocale, lSecurityFlags, strAuthority, \
            pCtx, ppNamespace = argv
        argv[1] = self.read_wide_string(strNetworkResource)

        if ppNamespace:
            ci = emu.com.get_interface(emu, emu.get_ptr_size(), 'IWbemServices')
            pNamespace = self.mem_alloc(emu.get_ptr_size(),
                                        tag='emu.COM.ppNamespace_IWbemServices')
            self.mem_write(pNamespace, ci.address.to_bytes(emu.get_ptr_size(), 'little'))
            self.mem_write(ppNamespace, pNamespace.to_bytes(emu.get_ptr_size(), 'little'))

        return comdefs.S_OK

    @apihook('IWbemServices.ExecQuery', argc=6)
    def IWbemServices_ExecQuery(self, emu, argv, ctx={}):
        """
        HRESULT ExecQuery(
            const BSTR           strQueryLanguage,
            const BSTR           strQuery,
            long                 lFlags,
            IWbemContext         *pCtx,
            IEnumWbemClassObject **ppEnum
        );
        """
        ptr, strQueryLanguage, strQuery, lFlags, pCtx, ppEnum = argv
        argv[1] = self.read_wide_string(strQueryLanguage)
        argv[2] = self.read_wide_string(strQuery)

        # not implemented so returning -1
        return -1



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.


from .. import api


class Comctl32(api.ApiHandler):

    name = 'comctl32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Comctl32, self).__init__(emu)
        self.funcs = {}
        self.data = {}
        super(Comctl32, self).__get_hook_attrs__(self)
        self.netman = emu.get_network_manager()
        self.names = {}

    @apihook('InitCommonControlsEx', argc=1)
    def InitCommonControlsEx(self, emu, argv, ctx={}):
        """
        BOOL InitCommonControlsEx(
            const INITCOMMONCONTROLSEX *picce
        );
        """
        picce, = argv
        rv = True

        return rv



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import os
import ntpath

from .. import api
import speakeasy.winenv.arch as e_arch


class Shlwapi(api.ApiHandler):

    """
    Implements exported functions from shlwapi.dll
    """

    name = 'shlwapi'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Shlwapi, self).__init__(emu)

        self.funcs = {}
        self.data = {}
        self.window_hooks = {}
        self.handle = 0
        self.win = None

        super(Shlwapi, self).__get_hook_attrs__(self)

    def join_windows_path(self, *args, **kwargs):
        args = list(map(lambda x: x.replace('\\', '/'), args))
        return os.path.join(*args, **kwargs).replace('/', '\\')

    @apihook('PathIsRelative', argc=1)
    def PathIsRelative(self, emu, argv, ctx={}):
        '''
        BOOL PathIsRelativeA(
            LPCSTR pszPath
        );
        '''

        pszPath, = argv

        cw = self.get_char_width(ctx)
        pn = ''
        rv = False
        if pszPath:
            pn = self.read_mem_string(pszPath, cw)
            if '..' in pn:
                rv = True

            argv[0] = pn

        return rv

    @apihook('StrStrI', argc=2)
    def StrStrI(self, emu, argv, ctx={}):
        '''
        PCSTR StrStrI(
            PCSTR pszFirst,
            PCSTR pszSrch
        );
        '''

        hay, needle = argv

        cw = self.get_char_width(ctx)

        if hay:
            _hay = self.read_mem_string(hay, cw)
            argv[0] = _hay
            _hay = _hay.lower()

        if needle:
            needle = self.read_mem_string(needle, cw)
            argv[1] = needle
            needle = needle.lower()

        ret = _hay.find(needle)
        if ret != -1:
            ret = hay + ret
        else:
            ret = 0

        return ret

    @apihook('PathFindExtension', argc=1)
    def PathFindExtension(self, emu, argv, ctx={}):
        """LPCSTR PathFindExtensionA(
          LPCSTR pszPath
        );
        """
        pszPath, = argv
        cw = self.get_char_width(ctx)
        s = self.read_mem_string(pszPath, cw)
        argv[0] = s
        idx1 = s.rfind('\\')
        t = s[idx1 + 1:]
        idx2 = t.rfind('.')
        if idx2 == -1:
            return pszPath + len(s)

        argv[0] = t[idx2:]
        return pszPath + idx1 + 1 + idx2

    @apihook('StrCmpI', argc=2)
    def StrCmpI(self, emu, argv, ctx={}):
        """
        int StrCmpI(
        PCWSTR psz1,
        PCWSTR psz2
        );
        """
        psz1, psz2 = argv

        cw = self.get_char_width(ctx)
        s1 = self.read_mem_string(psz1, cw)
        s2 = self.read_mem_string(psz2, cw)
        rv = 1

        argv[0] = s1
        argv[1] = s2

        if s1.lower() == s2.lower():
            rv = 0

        return rv

    @apihook('PathFindFileName', argc=1)
    def PathFindFileName(self, emu, argv, ctx={}):
        """
        LPCSTR PathFindFileNameA(
          LPCSTR pszPath
        );
        """
        pszPath, = argv
        cw = self.get_char_width(ctx)
        s = self.read_mem_string(pszPath, cw)
        argv[0] = s
        idx = s.rfind('\\')
        if idx == -1:
            return pszPath + len(s)

        argv[0] = s[idx + 1:]
        return pszPath + idx + 1

    @apihook('PathRemoveExtension', argc=1)
    def PathRemoveExtension(self, emu, argv, ctx={}):
        """
        void PathRemoveExtensionA(
          LPSTR pszPath
        );
        """
        pszPath, = argv
        cw = self.get_char_width(ctx)
        s = self.read_mem_string(pszPath, cw)
        argv[0] = s
        idx1 = s.rfind('\\')
        t = s[idx1 + 1:]
        idx2 = t.rfind('.')
        if idx2 == -1:
            return pszPath

        s = s[:idx1 + 1 + idx2]
        argv[0] = s
        self.write_mem_string(s, pszPath, cw)
        return pszPath

    @apihook('PathStripPath', argc=1)
    def PathStripPath(self, emu, argv, ctx={}):
        """
        void PathStripPath(
        LPSTR pszPath
        );
        """
        pszPath, = argv
        cw = self.get_char_width(ctx)
        s = self.read_mem_string(pszPath, cw)
        argv[0] = s
        mod_name = ntpath.basename(s) + '\x00'

        enc = self.get_encoding(cw)
        mod_name = mod_name.encode(enc)
        self.mem_write(pszPath, mod_name)

    @apihook('wvnsprintfA', argc=4)
    def wvnsprintfA(self, emu, argv, ctx={}):
        """
        int wvnsprintfA(
            PSTR    pszDest,
            int     cchDest,
            PCSTR   pszFmt,
            va_list arglist
        );
        """
        buffer, count, _format, argptr = argv
        rv = 0

        fmt_str = self.read_mem_string(_format, 1)
        fmt_cnt = self.get_va_arg_count(fmt_str)

        vargs = self.va_args(argptr, fmt_cnt)

        fin = self.do_str_format(fmt_str, vargs)
        fin = fin[:count] + '\x00'

        rv = len(fin)
        self.mem_write(buffer, fin.encode('utf-8'))
        argv[0] = fin.replace('\x00', '')
        argv[1] = fmt_str

        return rv

    @apihook('wnsprintf', argc=e_arch.VAR_ARGS, conv=e_arch.CALL_CONV_CDECL)
    def wnsprintf(self, emu, argv, ctx={}):
        """
        int wnsprintfA(
          PSTR  pszDest,
          int   cchDest,
          PCSTR pszFmt,
          ...
        );
        """
        argv = emu.get_func_argv(e_arch.CALL_CONV_CDECL, 3)
        buf, max_buf_size, fmt = argv

        cw = self.get_char_width(ctx)

        fmt_str = self.read_mem_string(fmt, cw)
        fmt_cnt = self.get_va_arg_count(fmt_str)
        if not fmt_cnt:
            self.write_mem_string(fmt_str, buf, cw)
            return len(fmt_str)

        _argv = emu.get_func_argv(e_arch.CALL_CONV_CDECL, 3 + fmt_cnt)[3:]
        fin = self.do_str_format(fmt_str, _argv)
        rv = len(fin)

        if rv <= max_buf_size:
            self.write_mem_string(fin, buf, cw)
            argv[0] = fin
            argv[2] = fmt_str
            return rv
        else:
            return -1

    @apihook('PathAppend', argc=2)
    def PathAppend(self, emu, argv, ctx={}):
        """
        BOOL PathAppendA(
          LPSTR  pszPath,
          LPCSTR pszMore
        );
        """
        pszPath, pszMore = argv
        cw = self.get_char_width(ctx)
        path = self.read_mem_string(pszPath, cw)
        more = self.read_mem_string(pszMore, cw)
        argv[0] = path
        argv[1] = more
        out = self.join_windows_path(path, more)
        out += '\0'
        self.write_mem_string(out, pszPath, cw)
        return 1



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from speakeasy.struct import EmuStruct, Ptr
import ctypes as ct


class WSK_PROVIDER_BASIC_DISPATCH(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.WskControlSocket = Ptr
        self.WskCloseSocket = Ptr


class WSK_PROVIDER_DATAGRAM_DISPATCH(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Basic = WSK_PROVIDER_BASIC_DISPATCH
        self.WskBind = Ptr
        self.WskSendTo = Ptr
        self.WskReceiveFrom = Ptr
        self.WskRelease = Ptr
        self.WskGetLocalAddress = Ptr
        self.WskSendMessages = Ptr


class WSK_CLIENT_DISPATCH(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Version = ct.c_uint16
        self.Reserved = ct.c_uint16
        self.WskClientEvent = Ptr


class WSK_CLIENT_NPI(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.ClientContext = Ptr
        self.Dispatch = Ptr


class WSK_PROVIDER_DISPATCH(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Version = ct.c_uint16
        self.Reserved = ct.c_uint16
        self.WskSocket = Ptr
        self.WskSocketConnect = Ptr
        self.WskControlClient = Ptr
        self.WskGetAddressInfo = Ptr
        self.WskFreeAddressInfo = Ptr
        self.WskGetNameInfo = Ptr


class WSK_PROVIDER_NPI(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Client = Ptr
        self.Dispatch = Ptr



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.arch as _arch
import speakeasy.winenv.defs.nt.ntoskrnl as w
from speakeasy.winenv.api import api


class Hal(api.ApiHandler):
    """
    Implements the hardware abstraction layer (hal.dll) that allows
    Windows to interact with hardware at a high level.
    """

    name = 'hal'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Hal, self).__init__(emu)

        self.funcs = {}
        self.data = {}
        self.emu = emu

        self.win = w

        super(Hal, self).__get_hook_attrs__(self)

    @apihook('KeGetCurrentIrql', argc=0)
    def KeGetCurrentIrql(self, emu, argv, ctx={}):
        """
        NTHALAPI KIRQL KeGetCurrentIrql();
        """
        irql = emu.get_current_irql()
        return irql

    @apihook('ExAcquireFastMutex', argc=1, conv=_arch.CALL_CONV_FASTCALL)
    def ExAcquireFastMutex(self, emu, argv, ctx={}):
        """
        VOID ExAcquireFastMutex(
            _Inout_ PFAST_MUTEX FastMutex
        );
        """
        return

    @apihook('ExReleaseFastMutex', argc=1, conv=_arch.CALL_CONV_FASTCALL)
    def ExReleaseFastMutex(self, emu, argv, ctx={}):
        """
        VOID ExReleaseFastMutex(
            _Inout_ PFAST_MUTEX FastMutex
        );
        """
        return



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

class CryptKey(object):
    def __init__(self, blob_type, blob, blob_len, hnd_import_key, param_list, flags):
        self.blob_type = blob_type
        self.blob = blob
        self.blob_len = blob_len
        self.import_key = hnd_import_key
        self.param_list = param_list
        self.flags = flags


class CryptContext(object):
    """
    Represents crypto context used by crypto functions
    """
    curr_handle = 0x680

    def __init__(self, cname, pname, ptype, flags):
        self.container_name = cname
        self.provider_name = pname
        self.ptype = ptype
        self.flags = flags
        self.keys = {}

    def get_handle(self):
        hkey = CryptContext.curr_handle
        CryptContext.curr_handle += 4
        return hkey

    def import_key(self, blob_type=None, blob=None, blob_len=None, hnd_import_key=None,
                   param_list=None, flags=None):
        key = CryptKey(blob_type, blob, blob_len, hnd_import_key, param_list, flags)
        hnd = self.get_handle()
        self.keys.update({hnd: key})

        return hnd

    def get_key(self, hnd):
        return self.keys.get(hnd, None)

    def delete_key(self, hnd):
        self.keys.pop(hnd)


class CryptoManager(object):
    """
    Manages the emulation of crypto functions
    """
    def __init__(self, config=None):
        super(CryptoManager, self).__init__()
        self.ctx_handles = {}
        self.config = config

    def crypt_open(self, cname=None, pname=None, ptype=None, flags=None):
        ctx = CryptContext(cname, pname, ptype, flags)
        hnd = ctx.get_handle()

        self.ctx_handles.update({hnd: ctx})
        return hnd

    def crypt_close(self, hnd):
        self.ctx_handles.pop(hnd)

    def crypt_get(self, hnd):
        return self.ctx_handles.get(hnd, None)



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import ctypes as ct
import speakeasy.winenv.arch as _arch
from speakeasy.struct import EmuStruct, Ptr

from .. import api


class WTS_SESSION_INFO(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.SessionId = ct.c_uint32
        self.pWinStationName = Ptr
        self.State = ct.c_uint32


class WtsApi32(api.ApiHandler):

    name = 'wtsapi32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(WtsApi32, self).__init__(emu)
        self.funcs = {}
        self.data = {}
        super(WtsApi32, self).__get_hook_attrs__(self)

    @apihook('WTSEnumerateSessions', argc=5, conv=_arch.CALL_CONV_STDCALL)
    def WTSEnumerateSessions(self, emu, argv, ctx={}):
        """
        BOOL WTSEnumerateSessions(
          IN HANDLE          hServer,
          IN DWORD           Reserved,
          IN DWORD           Version,
          PWTS_SESSION_INFO *ppSessionInfo,
          DWORD              *pCount
        );
        """

        hServer, res, ver, ppSessionInfo, pCount = argv
        rv = 0

        fn = ctx['func_name']
        cw = self.get_char_width(ctx)

        winstatname = 'RDP-Tcp#1' + '\x00'
        if cw == 2:
            sn = winstatname.encode('utf-16le')
        elif cw == 1:
            sn = winstatname.encode('utf-8')

        wsi = WTS_SESSION_INFO(emu.get_ptr_size())

        size = len(sn) + wsi.sizeof()
        buf = self.mem_alloc(size=size, tag='api.%s' % (fn))

        wsi.SessionId = 1
        # Write the string at the end of the structure
        wsi.pWinStationName = buf + wsi.sizeof()
        wsi.State = 0

        # Write the structure into memory
        self.mem_write(buf, self.get_bytes(wsi))
        # Write the string at the end of the structure
        self.mem_write(buf + wsi.sizeof(), sn)

        # Write the total session count
        if ppSessionInfo and pCount:
            self.mem_write(pCount, (1).to_bytes(4, 'little'))
            # Write the session buffer
            self.mem_write(ppSessionInfo, (buf).to_bytes(self.get_ptr_size(), 'little'))

            rv = 1

        return rv

    @apihook('WTSFreeMemory', argc=1, conv=_arch.CALL_CONV_STDCALL)
    def WTSFreeMemory(self, emu, argv, ctx={}):
        """
        void WTSFreeMemory(
          IN PVOID pMemory
        );
        """
        pMemory, = argv
        rv = 1

        self.mem_free(pMemory)

        return rv



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from socket import inet_aton
from urllib.parse import urlparse
import speakeasy.winenv.arch as _arch

import speakeasy.windows.netman as netman
import speakeasy.winenv.defs.wininet as windefs

from .. import api


def is_ip_address(ip):
    try:
        inet_aton(ip)
        return True
    except Exception:
        return False


class Wininet(api.ApiHandler):

    """
    Implements network functions from wininet.dll
    """

    name = 'wininet'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Wininet, self).__init__(emu)

        self.funcs = {}
        self.data = {}
        self.win = None
        self.netman = netman.NetworkManager(config=emu.get_network_config())
        super(Wininet, self).__get_hook_attrs__(self)

    @apihook('InternetOpen', argc=5, conv=_arch.CALL_CONV_STDCALL)
    def InternetOpen(self, emu, argv, ctx={}):
        """
        void InternetOpenA(
          LPTSTR lpszAgent,
          DWORD  dwAccessType,
          LPTSTR lpszProxy,
          LPTSTR lpszProxyBypass,
          DWORD  dwFlags
        );
        """
        ua, access, proxy, bypass, flags = argv

        cw = self.get_char_width(ctx)
        if ua:
            ua = self.read_mem_string(ua, cw)
            argv[0] = ua
        if proxy:
            proxy = self.read_mem_string(proxy, cw)
            argv[2] = proxy
        if bypass:
            bypass = self.read_mem_string(bypass, cw)
            argv[3] = bypass

        conn = self.netman.new_wininet_inst(ua, access, proxy, bypass, flags)
        hnd = conn.get_handle()
        return hnd

    @apihook('InternetConnect', argc=8, conv=_arch.CALL_CONV_STDCALL)
    def InternetConnect(self, emu, argv, ctx={}):
        """
        void InternetConnect(
          HINTERNET     hInternet,
          LPTSTR        lpszServerName,
          INTERNET_PORT nServerPort,
          LPTSTR        lpszUserName,
          LPTSTR        lpszPassword,
          DWORD         dwService,
          DWORD         dwFlags,
          DWORD_PTR     dwContext
        );
        """
        hnd, server, port, user, password, service, flags, dwctx = argv

        cw = self.get_char_width(ctx)
        if server:
            server = self.read_mem_string(server, cw)
            argv[1] = server
        if user:
            user = self.read_mem_string(user, cw)
            argv[3] = user
        if password:
            password = self.read_mem_string(password, cw)
            argv[4] = password

        wini = self.netman.get_wininet_object(hnd)

        if not wini:
            return 0

        sess = wini.new_session(server, port, user, password,
                                service, flags, dwctx)
        hdl = sess.get_handle()
        return hdl

    @apihook('HttpOpenRequest', argc=8, conv=_arch.CALL_CONV_STDCALL)
    def HttpOpenRequest(self, emu, argv, ctx={}):
        """
        void HttpOpenRequest(
          HINTERNET hConnect,
          LPTSTR    lpszVerb,
          LPTSTR    lpszObjectName,
          LPTSTR    lpszVersion,
          LPTSTR    lpszReferrer,
          LPTSTR    *lplpszAcceptTypes,
          DWORD     dwFlags,
          DWORD_PTR dwContext
        );
        """
        hnd, verb, objname, ver, ref, accepts, flags, dwctx = argv

        cw = self.get_char_width(ctx)
        if verb:
            verb = self.read_mem_string(verb, cw)
            argv[1] = verb
        if objname:
            objname = self.read_mem_string(objname, cw)
            argv[2] = objname
        if ver:
            ver = self.read_mem_string(ver, cw)
            argv[3] = ver
        if ref:
            ref = self.read_mem_string(ref, cw)
            argv[4] = ref

        defs = windefs.get_flag_defines(flags)
        argv[6] = ' | '.join(defs)

        sess = self.netman.get_wininet_object(hnd)
        req = sess.new_request(verb, objname, ver, ref, accepts, defs, dwctx)
        hdl = req.get_handle()
        return hdl

    @apihook('InternetCrackUrl', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def InternetCrackUrl(self, emu, argv, ctx={}):
        """
        BOOLAPI InternetCrackUrl(
            LPCSTR            lpszUrl,
            DWORD             dwUrlLength,
            DWORD             dwFlags,
            LPURL_COMPONENTSA lpUrlComponents
        );
        """
        lpszUrl, dwUrlLength, dwFlags, lpUrlComponents = argv

        rv = False
        cw = self.get_char_width(ctx)

        if lpszUrl and lpUrlComponents:
            url = self.read_mem_string(lpszUrl, cw)
            argv[0] = url
            rv = True

            uc = windefs.URL_COMPONENTS(emu.get_ptr_size())
            url_comp = self.mem_cast(uc, lpUrlComponents)

            crack = urlparse(url)
            if crack.scheme == 'https':
                url_comp.nScheme = windefs.INTERNET_SCHEME_HTTPS
            elif crack.scheme == 'http':
                url_comp.nScheme = windefs.INTERNET_SCHEME_HTTP
            if url_comp.dwHostNameLength > 0:
                if url_comp.lpszHostName:
                    host = crack.netloc + '\x00'
                    enc = self.get_encoding(cw)
                    self.mem_write(url_comp.lpszHostName, host.encode(enc))
                else:
                    offset = url.find(crack.netloc)
                    ptr = lpszUrl + (offset * cw)
                    url_comp.lpszHostName = ptr
                    url_comp.dwHostNameLength = len(crack.netloc)

            self.mem_write(lpUrlComponents, url_comp.get_bytes())

        return rv

    @apihook('InternetSetOption', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def InternetSetOption(self, emu, argv, ctx={}):
        """
        BOOLAPI InternetSetOption(
          HINTERNET hInternet,
          DWORD     dwOption,
          LPVOID    lpBuffer,
          DWORD     dwBufferLength
        );
        """
        hnd, option, buf, length = argv

        rv = 1

        return rv

    @apihook('HttpSendRequest', argc=5, conv=_arch.CALL_CONV_STDCALL)
    def HttpSendRequest(self, emu, argv, ctx={}):
        """
        BOOLAPI HttpSendRequest(
          HINTERNET hRequest,
          LPTSTR    lpszHeaders,
          DWORD     dwHeadersLength,
          LPVOID    lpOptional,
          DWORD     dwOptionalLength
        );
        """
        hnd, headers, hdrlen, lpOptional, dwOptionalLength = argv

        body = b''

        cw = self.get_char_width(ctx)
        if headers:
            headers = self.read_mem_string(headers, cw)
            argv[1] = headers

        if lpOptional:
            body = self.mem_read(lpOptional, dwOptionalLength)

        req = self.netman.get_wininet_object(hnd)
        srv = req.get_server()
        port = req.get_port()

        if not is_ip_address(srv):
            self.log_dns(srv)

        rv = 1
        req_str = req.format_http_request(headers=headers)

        self.log_http(srv, port, headers=req_str,
                      body=body, secure=req.is_secure())
        return rv

    @apihook('InternetErrorDlg', argc=5, conv=_arch.CALL_CONV_STDCALL)
    def InternetErrorDlg(self, emu, argv, ctx={}):
        """
        void InternetErrorDlg(
          HWND      hWnd,
          HINTERNET hRequest,
          DWORD     dwError,
          DWORD     dwFlags,
          LPVOID    *lppvData
        );
        """
        hWnd, req, error, flags, data = argv

        return

    @apihook('InternetQueryOption', argc=4)
    def InternetQueryOption(self, emu, argv, ctx={}):
        """
        BOOLAPI InternetQueryOption(
            HINTERNET hInternet,
            DWORD     dwOption,
            LPVOID    lpBuffer,
            LPDWORD   lpdwBufferLength
        );
        """
        hInternet, dwOption, lpBuffer, lpdwBufferLength = argv
        rv = False
        opt = windefs.get_option_define(dwOption)
        if opt:
            argv[1] = opt

        if dwOption == windefs.INTERNET_OPTION_SECURITY_FLAGS:
            if lpBuffer:
                sec_flags = windefs.SECURITY_FLAG_SECURE
                self.mem_write(lpBuffer, (sec_flags).to_bytes(4, 'little'))
                rv = True

        return rv

    @apihook('InternetReadFile', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def InternetReadFile(self, emu, argv, ctx={}):
        """
        BOOLAPI InternetReadFile(
          HINTERNET hFile,
          LPVOID    lpBuffer,
          DWORD     dwNumberOfBytesToRead,
          LPDWORD   lpdwNumberOfBytesRead
        );
        """
        hFile, buf, size, bytes_read = argv

        rv = 1

        req = self.netman.get_wininet_object(hFile)
        resp = req.get_response()
        data = resp.read(size)

        if buf:
            self.mem_write(buf, data)

        if bytes_read:
            self.mem_write(bytes_read, (len(data)).to_bytes(4, 'little'))

        return rv

    @apihook('HttpQueryInfo', argc=5)
    def HttpQueryInfo(self, emu, argv, ctx={}):
        """
        BOOLAPI HttpQueryInfo(
            HINTERNET hRequest,
            DWORD     dwInfoLevel,
            LPVOID    lpBuffer,
            LPDWORD   lpdwBufferLength,
            LPDWORD   lpdwIndex
        );
        """
        hRequest, dwInfoLevel, lpBuffer, lpdwBufferLength, lpdwIndex = argv
        cw = self.get_char_width(ctx)

        rv = False
        info_str = windefs.get_header_info(dwInfoLevel)
        if info_str:
            argv[1] = info_str
        if windefs.HTTP_QUERY_STATUS_CODE == dwInfoLevel:
            if lpBuffer:
                buf_len = self.mem_read(lpdwBufferLength, 4)
                buf_len = int.from_bytes(buf_len, 'little')

                if cw == 2:
                    enc = 'utf-16le'
                elif cw == 1:
                    enc = 'utf-8'
                out = windefs.HTTP_STATUS_OK.encode(enc)
                self.mem_write(lpBuffer, out)
                rv = True

        return rv

    @apihook('InternetQueryDataAvailable', argc=4)
    def InternetQueryDataAvailable(self, emu, argv, ctx={}):
        """
        BOOLAPI InternetQueryDataAvailable(
            HINTERNET hFile,
            LPDWORD   lpdwNumberOfBytesAvailable,
            DWORD     dwFlags,
            DWORD_PTR dwContext
        );
        """
        hFile, lpdwNumberOfBytesAvailable, dwFlags, dwContext = argv
        rv = False

        req = self.netman.get_wininet_object(hFile)
        avail = req.get_response_size()

        if lpdwNumberOfBytesAvailable:
            self.mem_write(lpdwNumberOfBytesAvailable, (avail.to_bytes(4, 'little')))
            rv = True

        return rv

    @apihook('InternetCloseHandle', argc=1)
    def InternetCloseHandle(self, emu, argv, ctx={}):
        """
        BOOLAPI InternetCloseHandle(
            HINTERNET hInternet
        );
        """
        hInternet, = argv
        rv = True

        self.netman.close_wininet_object(hInternet)

        return rv

    @apihook('InternetOpenUrl', argc=6)
    def InternetOpenUrl(self, emu, argv, ctx={}):
        """
        void InternetOpenUrlA(
            HINTERNET hInternet,
            LPCSTR    lpszUrl,
            LPCSTR    lpszHeaders,
            DWORD     dwHeadersLength,
            DWORD     dwFlags,
            DWORD_PTR dwContext
        );
        """
        hInternet, lpszUrl, lpszHeaders, dwHeadersLength, dwFlags, dwContext = argv
        cw = self.get_char_width(ctx)
        if lpszUrl:
            url = self.read_mem_string(lpszUrl, cw)
            argv[1] = url
        if lpszHeaders:
            headers = self.read_mem_string(lpszHeaders, cw)
            argv[2] = headers

        defs = windefs.get_flag_defines(dwFlags)
        argv[4] = ' | '.join(defs)

        wini = self.netman.get_wininet_object(hInternet)
        if not wini:
            return 0
        crack = urlparse(url)
        if crack.scheme == "http":
            # FIXME : parse port in url netloc
            port = 80
        else:
            port = 443
        self.log_http(crack.netloc, port, headers=lpszHeaders)
        sess = wini.new_session(crack.netloc, port, '', '', '', defs, dwContext)
        if not sess:
            return 0
        req = sess.new_request("GET", url, None, None, None, defs, dwContext)
        return req.get_handle()



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import os
import json
import lzma
import multiprocessing as mp

from speakeasy import Speakeasy


def get_test_bin_path(bin_name):
    fp = os.path.join(os.path.dirname(__file__), 'bins', bin_name)
    return fp


def get_test_bin_data(bin_name):
    fp = get_test_bin_path(bin_name)
    with lzma.open(fp) as f:
        file_content = f.read()
    return file_content


def get_config():
    fp = os.path.join(os.path.dirname(__file__), 'test.json')
    with open(fp, 'r') as f:
        return json.load(f)


def _run_module(q, cfg, target, logger, argv):
    se = Speakeasy(config=cfg, logger=logger, argv=argv)
    module = se.load_module(data=target)
    se.run_module(module, all_entrypoints=True)
    report = se.get_report()
    q.put(report)


def run_test(cfg, target, logger=None, argv=[]):

    q = mp.Queue()
    p = mp.Process(target=_run_module, args=(q, cfg, target, logger, argv))
    p.start()
    return q.get()



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import ctypes as ct
from speakeasy.struct import EmuStruct, Ptr


INTERNET_FLAG_ASYNC = 0x10000000
INTERNET_FLAG_CACHE_ASYNC = 0x00000080
INTERNET_FLAG_CACHE_IF_NET_FAIL = 0x00010000
INTERNET_FLAG_DONT_CACHE = 0x04000000
INTERNET_FLAG_EXISTING_CONNECT = 0x20000000
INTERNET_FLAG_FORMS_SUBMIT = 0x00000040
INTERNET_FLAG_FROM_CACHE = 0x01000000
INTERNET_FLAG_FWD_BACK = 0x00000020
INTERNET_FLAG_HYPERLINK = 0x00000400
INTERNET_FLAG_IGNORE_CERT_CN_INVALID = 0x00001000
INTERNET_FLAG_IGNORE_CERT_DATE_INVALID = 0x00002000
INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTP = 0x00008000
INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTPS = 0x00004000
INTERNET_FLAG_KEEP_CONNECTION = 0x00400000
INTERNET_FLAG_MAKE_PERSISTENT = 0x02000000
INTERNET_FLAG_MUST_CACHE_REQUEST = 0x00000010
INTERNET_FLAG_NEED_FILE = 0x00000010
INTERNET_FLAG_NO_AUTH = 0x00040000
INTERNET_FLAG_NO_AUTO_REDIRECT = 0x00200000
INTERNET_FLAG_NO_COOKIES = 0x00080000
INTERNET_FLAG_NO_UI = 0x00000200
INTERNET_FLAG_OFFLINE = 0x01000000
INTERNET_FLAG_PASSIVE = 0x08000000
INTERNET_FLAG_PRAGMA_NOCACHE = 0x00000100
INTERNET_FLAG_RAW_DATA = 0x40000000
INTERNET_FLAG_READ_PREFETCH = 0x00100000
INTERNET_FLAG_RELOAD = 0x80000000
INTERNET_FLAG_RESTRICTED_ZONE = 0x00020000
INTERNET_FLAG_RESYNCHRONIZE = 0x00000800
INTERNET_FLAG_SECURE = 0x00800000
INTERNET_FLAG_TRANSFER_ASCII = 0x00000001
INTERNET_FLAG_TRANSFER_BINARY = 0x00000002
INTERNET_NO_CALLBACK = 0x00000000
INTERNET_OPTION_SUPPRESS_SERVER_AUTH = 104
WININET_API_FLAG_ASYNC = 0x00000001
WININET_API_FLAG_SYNC = 0x00000004
WININET_API_FLAG_USE_CONTEXT = 0x00000008

INTERNET_SCHEME_PARTIAL = -2
INTERNET_SCHEME_UNKNOWN = -1
INTERNET_SCHEME_DEFAULT = 0
INTERNET_SCHEME_FTP = 1
INTERNET_SCHEME_GOPHER = 2
INTERNET_SCHEME_HTTP = 3
INTERNET_SCHEME_HTTPS = 4

HTTP_QUERY_STATUS_CODE = 19
HTTP_STATUS_OK = "200"

WINHTTP_ADDREQ_INDEX_MASK = 0x0000FFFF
WINHTTP_ADDREQ_FLAGS_MASK = 0xFFFF0000
WINHTTP_ADDREQ_FLAG_ADD_IF_NEW = 0x10000000
WINHTTP_ADDREQ_FLAG_ADD = 0x20000000
WINHTTP_ADDREQ_FLAG_COALESCE_WITH_COMMA = 0x40000000
WINHTTP_ADDREQ_FLAG_COALESCE_WITH_SEMICOLON = 0x01000000

WINHTTP_QUERY_MIME_VERSION = 0
WINHTTP_QUERY_CONTENT_TYPE = 1
WINHTTP_QUERY_CONTENT_TRANSFER_ENCODING = 2
WINHTTP_QUERY_CONTENT_ID = 3
WINHTTP_QUERY_CONTENT_DESCRIPTION = 4
WINHTTP_QUERY_CONTENT_LENGTH = 5
WINHTTP_QUERY_CONTENT_LANGUAGE = 6
WINHTTP_QUERY_ALLOW = 7
WINHTTP_QUERY_PUBLIC = 8
WINHTTP_QUERY_DATE = 9
WINHTTP_QUERY_EXPIRES = 10
WINHTTP_QUERY_LAST_MODIFIED = 11
WINHTTP_QUERY_MESSAGE_ID = 12
WINHTTP_QUERY_URI = 13
WINHTTP_QUERY_DERIVED_FROM = 14
WINHTTP_QUERY_COST = 15
WINHTTP_QUERY_LINK = 16
WINHTTP_QUERY_PRAGMA = 17
WINHTTP_QUERY_VERSION = 18
WINHTTP_QUERY_STATUS_CODE = 19
WINHTTP_QUERY_STATUS_TEXT = 20
WINHTTP_QUERY_RAW_HEADERS = 21
WINHTTP_QUERY_RAW_HEADERS_CRLF = 22
WINHTTP_QUERY_CONNECTION = 23
WINHTTP_QUERY_ACCEPT = 24
WINHTTP_QUERY_ACCEPT_CHARSET = 25
WINHTTP_QUERY_ACCEPT_ENCODING = 26
WINHTTP_QUERY_ACCEPT_LANGUAGE = 27
WINHTTP_QUERY_AUTHORIZATION = 28
WINHTTP_QUERY_CONTENT_ENCODING = 29
WINHTTP_QUERY_FORWARDED = 30
WINHTTP_QUERY_FROM = 31
WINHTTP_QUERY_IF_MODIFIED_SINCE = 32
WINHTTP_QUERY_LOCATION = 33
WINHTTP_QUERY_ORIG_URI = 34
WINHTTP_QUERY_REFERER = 35
WINHTTP_QUERY_RETRY_AFTER = 36
WINHTTP_QUERY_SERVER = 37
WINHTTP_QUERY_TITLE = 38
WINHTTP_QUERY_USER_AGENT = 39
WINHTTP_QUERY_WWW_AUTHENTICATE = 40
WINHTTP_QUERY_PROXY_AUTHENTICATE = 41
WINHTTP_QUERY_ACCEPT_RANGES = 42
WINHTTP_QUERY_SET_COOKIE = 43
WINHTTP_QUERY_COOKIE = 44
WINHTTP_QUERY_REQUEST_METHOD = 45
WINHTTP_QUERY_REFRESH = 46
WINHTTP_QUERY_CONTENT_DISPOSITION = 47
WINHTTP_QUERY_AGE = 48
WINHTTP_QUERY_CACHE_CONTROL = 49
WINHTTP_QUERY_CONTENT_BASE = 50
WINHTTP_QUERY_CONTENT_LOCATION = 51
WINHTTP_QUERY_CONTENT_MD5 = 52
WINHTTP_QUERY_CONTENT_RANGE = 53
WINHTTP_QUERY_ETAG = 54
WINHTTP_QUERY_HOST = 55
WINHTTP_QUERY_IF_MATCH = 56
WINHTTP_QUERY_IF_NONE_MATCH = 57
WINHTTP_QUERY_IF_RANGE = 58
WINHTTP_QUERY_IF_UNMODIFIED_SINCE = 59
WINHTTP_QUERY_MAX_FORWARDS = 60
WINHTTP_QUERY_PROXY_AUTHORIZATION = 61
WINHTTP_QUERY_RANGE = 62
WINHTTP_QUERY_TRANSFER_ENCODING = 63
WINHTTP_QUERY_UPGRADE = 64
WINHTTP_QUERY_VARY = 65
WINHTTP_QUERY_VIA = 66
WINHTTP_QUERY_WARNING = 67
WINHTTP_QUERY_EXPECT = 68
WINHTTP_QUERY_PROXY_CONNECTION = 69
WINHTTP_QUERY_UNLESS_MODIFIED_SINCE = 70
WINHTTP_QUERY_PROXY_SUPPORT = 75
WINHTTP_QUERY_AUTHENTICATION_INFO = 76
WINHTTP_QUERY_PASSPORT_URLS = 77
WINHTTP_QUERY_PASSPORT_CONFIG = 78
WINHTTP_QUERY_MAX = 78


INTERNET_OPTION_SECURITY_FLAGS = 31

SECURITY_FLAG_SECURE = 1

ERROR_INSUFFICIENT_BUFFER = 122


class URL_COMPONENTS(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.dwStructSize = ct.c_uint32
        self.lpszScheme = Ptr
        self.dwSchemeLength = ct.c_uint32
        self.nScheme = ct.c_uint32
        self.lpszHostName = Ptr
        self.dwHostNameLength = ct.c_uint32
        self.nPort = ct.c_uint16
        self.lpszUserName = Ptr
        self.dwUserNameLength = ct.c_uint32
        self.lpszPassword = Ptr
        self.dwPasswordLength = ct.c_uint32
        self.lpszUrlPath = Ptr
        self.dwUrlPathLength = ct.c_uint32
        self.lpszExtraInfo = Ptr
        self.dwExtraInfoLength = ct.c_uint32


def get_const_defines(flags, prefix=''):
    defs = []
    for k, v in globals().items():
        if isinstance(v, int):
            if v & flags:
                if prefix:
                    if k.startswith(prefix):
                        defs.append(k)
                else:
                    defs.append(k)
    return defs


def get_flag_defines(flags):
    return get_const_defines(flags, prefix='INTERNET_FLAG')


def get_header_info(info):
    for k, v in globals().items():
        if k.startswith('HTTP_QUERY') and v == info:
            return k


def get_option_define(opt):
    for k, v in globals().items():
        if k.startswith('INTERNET_OPTION_') and v == opt:
            return k


def get_header_info_winhttp(flags):
    return get_const_defines(flags, prefix='WINHTTP_ADDREQ_')


def get_header_query(opt):
    for k, v in globals().items():
        if k.startswith('WINHTTP_QUERY_') and v == opt:
            return k



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from speakeasy.struct import EmuStruct
import ctypes as ct


class POINT(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.x = ct.c_uint32
        self.y = ct.c_uint32


class RECT(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.left = ct.c_int32
        self.top = ct.c_int32
        self.right = ct.c_int32
        self.bottom = ct.c_int32


class MONITORINFO(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.cbSize = ct.c_uint32
        self.rcMonitor = RECT
        self.rcWORK = RECT
        self.dwFlags = ct.c_uint32
